import numpy as np
import pandas as pd
import scipy.stats as stat
import code
import copy
import sys
import itertools
from collections import Counter

from . import beliefs
from . import decision

class Agents():
    def __init__(self, inputs, market):
        # attribute the parameters to the object

        self.all_inputs = inputs
        self.inputs = inputs['agents']
        for key, val in self.inputs.items():
            setattr(self, key, val)

        self.N = self.all_inputs['model']['n_agents']
        self.T = self.all_inputs['model']['T']
        self.id = np.arange(self.N)
        self.savings_acct = bool(self.savings_acct)

        # generate land ownership (this will later be overwritten by data)
        self.land_area = self.init_farm_size().astype(float) # this represents cultivated area
        self.has_land = self.land_area>0
        self.land_rented = np.full(self.N, 0.) # assume no land rental. this is within the land_area (i.e., not additional to it)
        self.in_prospective_outgrower = np.full(self.N, False) # whether HH is physically in outgrower scheme BEFORE IT HAPPENS
        self.forced_outgrower = np.full(self.N, False) # whether HH is physically in outgrower scheme AFTER IT HAPPENS
        self.outgrower_yr = np.full([self.T,self.N], False) # annual
        self.crop_production = {}
        for crop in self.all_inputs['land']['ag_types']:
            self.crop_production[crop] = np.full([self.T, self.N], 0) # kg
        self.tot_crop_production = np.full([self.T, self.N], 0) # kg
        self.mkt_orientation = np.full([self.T, self.N], np.nan) # fraction of production that is sold
        self.self_sufficiency = np.full([self.T, self.N], np.nan) # fraction of food consumption that is self-produced
        # networks
        self.init_network()

        # savings and livestock
        # this represents the START of the year
        self.savings = np.full([self.T+1, self.N], -9999) # at start of year
        self.savings[0] = np.random.normal(self.savings_init_mean, self.savings_init_sd, self.N)
        self.savings[0][self.savings[0]<0] = 0 # fix any -ve values
        self.livestock = np.full([self.T+1, self.N], -9999) # at start of year
        self.livestock[0] = np.floor(self.livestock_init).astype(int) # constant amount for each agent (same as savings)
        self.wealth = np.full([self.T+1, self.N], -9999) # sum of livestock + savings. represents the value at the START of the year
        self.wealth[0] = self.savings[0] + self.livestock[0]*market.livestock_cost
        # money
        self.cash_income = np.full([self.T, self.N], np.nan) #
        self.cash_expenditure = np.full([self.T, self.N], -99999) #
        self.food_expenditure = np.full([self.T, self.N], -99999) #
        self.food_amt_sell_debts = np.full([self.T, self.N], -99999) #
        self.extra_spending = np.full([self.T, self.N], -99999) #
        # self.income = np.full([self.T, self.N], -9999) # for keeping track of net income
        # self.total_income = np.full([self.T, self.N], -9999)
        self.tot_farm_income = np.full([self.T, self.N], 0)
        # self.food_cost = np.full([self.T, self.N], 0)
        self.farm_income = {}
        self.farm_costs =np.full([self.T, self.N], 0)
        self.outgrower_costs =np.full([self.T, self.N], 0)
        for crop in self.all_inputs['land']['ag_types']:
            self.farm_income[crop] = np.full([self.T, self.N], 0)
        self.ls_income = np.full([self.T, self.N], 0)
        self.ls_sell_income = np.full([self.T, self.N], 0) # from selling livestock
        self.salary_income = np.full([self.T, self.N], -9999)
        self.wage_income = np.full([self.T, self.N], -9999)
        self.savings_post_cons_smooth = np.full([self.T, self.N], -9999)
        self.buy_food = np.full([self.T, self.N], False)
        self.ls_stocking_amt = np.full([self.T+1, self.N], 0) # amt of livestock that CAN be stocked
        self.ls_stocking_amt[0] = 0
        # coping measures
        self.food_shortage = np.full([self.T, self.N], False)
        self.reduce_food_intake = np.full([self.T, self.N], False)
        self.reduce_food_intake_amt = np.full([self.T, self.N], np.nan)
        self.cons_red_rqd = np.full([self.T, self.N], False)
        self.coping_rqd = np.full([self.T, self.N], False)
        self.neg_income = np.full([self.T, self.N], False)
        self.destocking_rqd = np.full([self.T, self.N], False)
        self.destocking_rqd_rangeland = np.full([self.T, self.N], False)
        self.stress_ls_sell_rqd = np.full([self.T, self.N], False)
        self.cant_cope = np.full([self.T, self.N], False)
        # other livestock values for record keeping
        self.ls_start = np.full([self.T, self.N], -9999) # before the decision-making
        self.ls_decision = np.full([self.T, self.N], -9999) # after the decision-making
        self.ls_reprod = np.full([self.T, self.N], -9999) # after reproduction
        self.ls_rangeland = np.full([self.T, self.N], -9999) # after rangeland dynamics
        self.ls_stress = np.full([self.T, self.N], -9999) # after stress selling
        self.ls_stocking = np.full([self.T, self.N], -9999) # after stocking
        self.max_ls_purchase = np.full([self.T, self.N], np.nan)
        val = -99 if self.all_inputs['rangeland']['integer_consumption'] else np.nan
        self.attempted_herds_on_rangeland = np.full([self.T, self.N], val)
        self.herds_on_rangeland = np.full([self.T, self.N], val)
        self.herds_on_residue = np.full([self.T, self.N], val)
        self.herds_on_others_residue = np.full([self.T, self.N], val)
        # labor
        self.n_yr_smooth = int(self.n_yr_smooth) # in case it's from POM
        self.ag_labor_rqmt = {'subs' : self.ag_labor_rqmt_subs, 'mkt' : self.ag_labor_rqmt_mkt}
        self.ag_labor = {}
        for crop in self.all_inputs['land']['ag_types']:
            self.ag_labor[crop] = np.full([self.T, self.N], np.nan)
        self.tot_ag_labor = np.full([self.T, self.N], 0.) # sum over all crops
        self.salary_labor = np.full([self.T, self.N], 0.) # must be zero. it looks at the final time step in the first step for the 'previous' assignment
        self.salary_consider = np.full([self.T, self.N], False)
        self.wage_labor = np.full([self.T, self.N], np.nan)
        self.ls_labor = np.full([self.T, self.N], np.nan)
        self.total_labor = np.full([self.T, self.N], 0.)
        self.salary_success = np.full([self.T, self.N], -99) # -99 means nan
        self.contract_honored = np.full([self.T, self.N], -99) # -99 means nan
        self.labor_density = np.full([self.T, self.N], np.nan) # represents labor availability / (ag labor demand)

        ##### DATA IMPORT #####
        # this overwrites the previous things if necessary
        if 'agents' in self.all_inputs['model']['read_from_file']:
            # init the households from the specified site/treatment pair
            self.init_from_file()
            self.file_init = True
        else:
            self.land_rented_post = self.land_rented
            self.in_prospective_outgrower = np.full(self.N, False)
            self.land_area_change = 0
            self.land_area_post = copy.deepcopy(self.land_area)
            self.land_area_post_cf = copy.deepcopy(self.land_area)
            self.file_init = False

        # living cost
        cost_value_site = self.living_cost_pp[self.all_inputs['model']['site']] if isinstance(self.living_cost_pp, dict) else self.living_cost_pp
        self.living_cost = self.hh_size * cost_value_site
        self.food_rqmt = self.food_rqmt_pp * self.hh_size

        # DECISIONS and BELIEFS
        self.init_beliefs()

    def init_beliefs(self):
        inp = self.all_inputs['decisions']
        # initialize beliefs
        self.blf = beliefs.Beliefs(self, self.all_inputs)
        # outgrower options (optional)
        outgrower_options = [0,1,-1] if self.all_inputs['adaptation']['outgrower']['join'] else [0] # decision to do nothing or participate
        if len(outgrower_options)>1:
            inp['actions']['outgrower'] = outgrower_options
        # initialize decisions
        self.decision_options = []
        keyz = list(inp['actions'].keys())
        combs = list(itertools.product(*list(inp['actions'].values())))
        for c in combs:
            tmp = {}
            for k, ki in enumerate(keyz):
                tmp[ki] = c[k]
            self.decision_options.append(tmp)
        self.choices = {}
        for act, vals in inp['actions'].items():
            fill_val = False if isinstance(vals[0],bool) else -99 # either bool or int
            self.choices[act] = np.full([self.T, self.N], fill_val)
        self.choice_ixs = np.full([self.T, self.N], -1)
        self.fallow = np.full([self.T+1, self.N], False)
        self.exp_util = np.full([self.T, len(self.decision_options), self.N], np.nan)
        self.pos_util = np.full([self.T, len(self.decision_options), self.N], np.nan)
        self.max_utility = np.full([self.T, self.N], np.nan)
        # decision-related agent properties
        self.option_feasibility = np.full((self.T, len(self.decision_options), self.N), False)
        # create dictionary of "similar" decisions for each decision option
        N_dec = len(self.decision_options)
        rng = range(N_dec)
        valz = np.array([list(self.decision_options[i].values()) for i in rng]).astype(int)
        sims = np.array([(valz == valz[i]).sum(1) for i in rng])
        sims2 = np.full([N_dec,N_dec],-1)
        boolz = [len(el)==2 for el in list(inp['actions'].values())]
        for i in rng:
            for j in rng:
                sum_ij = 0
                for e, el in enumerate(valz[i]):
                    if boolz[e]:
                        # increase the sum if it's the same
                        sum_ij += 1 if el==valz[j][e] else 0
                    else:
                        # increase the sum only if j==0 (ie "no change")
                        sum_ij += 1 if el==0 else 0
                sims2[i,j] = sum_ij # i am 90% sure the i and j are the right way around...
        self.decision_similarity = sims2

        if inp['risk_aversion']:
            rand_int = np.random.randint(1e6) # generate random integer to control stochasticity
            # high risk tolerance means more tolerant to risk
            self.risk_tolerance = np.random.normal(loc=inp['risk_tolerance_mu'], scale=inp['risk_tolerance_cov']*inp['risk_tolerance_mu'], size=self.N)
            self.risk_tolerance[self.risk_tolerance<1] = 1 # for instability in the exponential function
            # pre-generate the agent-level "realizations" from a standard normal distribution
            # that they will each use in their utility calculations
            self.rndm_Zs = np.random.normal(size=(self.T, self.N, inp['nsim_utility']))
            # generate random uniform numbers also -- these are not correlated to the normally distributed beliefs
            self.rndm_Us = np.random.uniform(size=(self.T, self.N, inp['nsim_utility']))
            np.random.seed(rand_int)

        self.discount_rate = inp['discount_rate']

    def init_from_file(self):
        '''
        initialize agent properties from the survey data
        if counterfactual=false, then use the specified site and treatment data
        else, parameterize the HHs to represent the counterfactual matched values (i.e. HHs from the control site matched to HHs in the treatment)
        if outgrower:
            identify the HHs that are "affected" - i.e. those that participate in the outgrower
        '''
        d_in = pd.read_csv(self.file_name, index_col=0)

        mod_inp = self.all_inputs['model']
        plot_size = self.all_inputs['land']['plot_size']
        rand_int = np.random.randint(1e6)
        site_list = ['OR1','OR2','OR4','OR5'] if mod_inp['site']=='ORX' else [mod_inp['site']]
        if mod_inp['matching_experiment']:
            if mod_inp['treat'] == 'Control':
                print('ERROR (agents): specified counterfactual with treat==Control')
                sys.exit()
            # read in the matches
            if mod_inp['use_weights']:
                # not made for ORx...
                w = pd.read_csv(mod_inp['hh_weights_file'].format(mod_inp['site']), index_col=0)
                trt_ids = w.loc[w['treat']==1,'HHID']
            else:
                d_cf = pd.read_csv(mod_inp['hh_match_file'], index_col=0)
                site_locs = d_cf['site'].isin(site_list)
                trt_ids = d_cf.loc[site_locs,'Treatment']
            
            # subset to this
            ilocs = [np.where(d_in['HHID']==i)[0][0] for i in trt_ids]
            d_in_subs = d_in.iloc[ilocs]
            # if mod_inp['LSLA_LUC']:
            #     d_in_subs_cf = d_in_subs
            # else:
            # identify the relevant counterfactual indices
            if mod_inp['use_weights']:
                # sample randomly using the weights -- this honors the HH weights, but not HH-level matching
                w_c = w.loc[w['treat']==0]
                cf_ids = np.random.choice(np.array(w_c['HHID']), p=w_c['weight']/w_c['weight'].sum(), size=trt_ids.shape[0], replace=True)
            else:
                # take the household-level matched counterfactuals
                cf_ids = d_cf.loc[site_locs, 'Control']

            ilocs_cf = [np.where(d_in['HHID']==i)[0][0] for i in cf_ids]
            d_in_subs_cf = d_in.iloc[ilocs_cf] # use these data for the "change" questions
        else:
            # filter simply by site and treatment
            d_in_subs = d_in.query('site in {} & treat=="{}" & hh_size>0'.format(site_list, mod_inp['treat']))
            d_in_subs_cf = d_in_subs # use same HHs for the "change" questions
        
        # set the counterfactual values
        post_qs = ['land_cultiv_change','land_cultiv_rent_change']
        for qi in post_qs:
            d_in_subs.insert(0,qi+'_cf', np.array(d_in_subs_cf[qi]))
        
        # remove rows with NAs
        d_in_subs = d_in_subs.loc[~(d_in_subs[['fs_days','land_cultiv_pre','ls_TLU_pre']].isna().sum(1)>0)]
        d_in_subs.index = np.arange(d_in_subs.shape[0])
        d_in_subs_cf.index = np.arange(d_in_subs_cf.shape[0])

        # define the sampling
        np.random.seed(rand_int)
        replace = False if self.N <= d_in_subs.shape[0] else True
        hh_ixs = np.random.choice(d_in_subs.shape[0], self.N, replace=replace)

        # loop over the variables
        land_rented_change = 0 # initialize this in case land_rented is not read from empirical data
        for el in self.props_from_file:
            ## adapt this for each variable added
            if el == 'hh_size':
                self.hh_size = np.array(d_in_subs.loc[hh_ixs,'hh_size'])
            elif el == 'land_area':
                self.land_area = round_down(np.array(d_in_subs.loc[hh_ixs, 'land_cultiv_pre']), plot_size)
                self.has_land = self.land_area>0
            elif el == 'land_rental':
                self.land_rented = round_down(np.array(d_in_subs.loc[hh_ixs, 'land_cultiv_rent_pre']), plot_size)
                if mod_inp['hh_lulc_data']=='survey':
                    cname = 'land_cultiv_rent_change' if self.all_inputs['LSLA']['LUC'] else 'land_cultiv_rent_change_cf'
                    land_rented_change = round_down(np.array(d_in_subs.loc[hh_ixs, cname]), plot_size)
            elif el == 'livestock':
                self.livestock[0] = round(d_in_subs.loc[hh_ixs, 'ls_TLU_pre'])
            elif el == 'decisions':
                tmp_nf = np.array(d_in_subs.loc[hh_ixs,'lbr_wage_pre'])
                tmp_nf[np.isnan(tmp_nf)] = 0
                tmp_nf = tmp_nf > 0
                self.decisions_init = {'fertilizer' : np.array(d_in_subs.loc[hh_ixs,'fert_bin_pre']),
                                        'salary_labor' : tmp_nf.astype(int),
                                        'ls_stock' : np.full(self.N, True), # assume they stock by default (although not actually valid at t==0)
                                        'outgrower' : np.full(self.N, 0),} # assume they don't try to join outgrower immediately
            else:
                print('ERROR (agents): Undefined empirical data parameter specified ({})'.format(el))
                sys.exit()

        ## simulate the changes in land
        self.land_rented_post = self.land_rented + land_rented_change
        self.land_cultivated_changes(mod_inp, d_in_subs, hh_ixs, plot_size)
        ## determine which HHs in outgrower if relevant
        self.hhs_in_outgrower(d_in_subs, hh_ixs)

        # checks
        if (self.land_rented > self.land_area).sum() > 0:
            print('ERROR (agents): land rented is more than land cultivated')
            sys.exit()
        if mod_inp['print_init']:
            print('  HHIDs : {} ...'.format(hh_ixs[0:5]))
            print('  land_change : {}'.format(cname))
        #### take final characteristics from the empirical data (for calibration) ####
        # take initial and final where possible
        df_post = d_in_subs_cf if (mod_inp['matching_experiment'] and mod_inp['matched_ids']) else d_in_subs # NOTE: NOT SURE ABOUT THIS....should i attribute yields from the counterfactual site? it probably doesn't make sense..
        self.id_empirical = df_post.loc[hh_ixs,'HHID']
        self.empirical = {}
        self.empirical['livestock'] = [np.array(np.round(d_in_subs.loc[hh_ixs,'ls_TLU_pre'])),np.array(np.round(df_post.loc[hh_ixs,'ls_TLU']))]
        self.empirical['fert_use'] = [np.array(d_in_subs.loc[hh_ixs,'fert_bin_pre']),np.array(df_post.loc[hh_ixs,'fert_bin'])]
        self.empirical['wage_labor'] = [np.array(d_in_subs.loc[hh_ixs,'lbr_wage_pre']),np.array(df_post.loc[hh_ixs,'lbr_wage'])]
        self.empirical['food_shortage_days'] = np.array(df_post.loc[hh_ixs,'fs_days'])
        self.empirical['crop_yield'] = np.array(df_post.loc[hh_ixs,'crop_yield'])
        self.empirical['land_cultiv_change'] = np.array(d_in_subs.loc[hh_ixs,'land_cultiv_change'])
        # code.interact(local=dict(globals(), **locals()))

    def hhs_in_outgrower(self, d_in_subs, hh_ixs):
        '''
        identify which HHs are in the outgrower (if relevant)
        '''
        # in outgrower if the land lost in Treatment is more than the counterfactual
        self.in_prospective_outgrower = self.land_area_post < self.land_area_post_cf
        self.forced_outgrower_ha = np.maximum(self.land_area_post_cf - self.land_area_post, 0) # this is rqd. only set up for this option so far (ie "land_lost")
        # by default can't join outgrower at the start
        self.can_join_outgrower = False

    def land_cultivated_changes(self, mod_inp, d_in_subs, hh_ixs, plot_size):
        '''
        simulate changes in land cultivated
        '''
        rand_int = np.random.randint(1e6) # generate random integer to control stochasticity
        if mod_inp['hh_lulc_data'] == 'survey':
            ### use the survey data to simulate changes in land
            ## LAND CULTIVATED
            # use the counterfactual value if relevant
            cname = 'land_cultiv_change' if self.all_inputs['LSLA']['LUC'] else 'land_cultiv_change_cf'
            self.land_area_change = round_down(np.array(d_in_subs.loc[hh_ixs, cname]), plot_size)
            # save the counterfactual value to use with outgrower scheme, if necessary
            self.land_area_post_cf = self.land_area + round_down(np.array(d_in_subs.loc[hh_ixs, 'land_cultiv_change_cf']), plot_size)
        elif mod_inp['hh_lulc_data'] == 'LULC':
            if not mod_inp['matching_experiment']:
                print('ERROR (agents): using LULC data outside of a matched experiment. code not designed for this')
                sys.exit()
            self.land_area_post_cf = copy.deepcopy(self.land_area) # assume no change for counterfactual/outgrower experiment
            # calculate land area changes
            if self.all_inputs['LSLA']['LUC']:
                self.land_area_change = np.full(self.N, 0.)
                # simulate based on the site-level LULCC 
                lc_data = pd.read_csv(mod_inp['LULC_file'], index_col=0)
                lc_site = lc_data[lc_data.site==mod_inp['site']]
                if mod_inp['site'] == 'ORX':
                    lc_site['sh_frac_lsla_buf'] = self.all_inputs['LSLA']['frac_smallholder']
                # if not outgrower, agents always lose all land
                area_loss = float(round_down(lc_site.area_lsla_frac * lc_site.sh_frac_lsla_buf * self.land_area.sum() / lc_site.sh_frac_buf_pre, plot_size)) # need to divide by fraction smallholder since the LSLA is measured as fraction of TOTAL area
                # if outgrower, less agents lose land (ie are involved) if it is only partially implemented
                area_loss *= self.all_inputs['LSLA']['implementation_fraction'] if self.all_inputs['LSLA']['outgrower'] else 1
                # qual_land_loss = np.array(d_in_subs.loc[hh_ixs, 'land_lost_qual']).astype(bool)
                if area_loss > 0:
                    # if self.land_area[qual_land_loss].sum() < area_loss:
                    #     print('ERROR (agents): too much land loss from LSLA')
                    #     sys.exit()
                    land_taking_type = self.all_inputs['LSLA']['land_taking_type']
                    if land_taking_type=='random_field':
                        # randomly select fields from these agents
                        field_owners = np.repeat(self.id, (self.land_area/plot_size).astype(int))
                        field_loss_ids = np.random.choice(field_owners, np.floor(area_loss/plot_size).astype(int), replace=False)
                        loss_counts = np.array(list(Counter(field_loss_ids).items()))
                        # assign to agents
                        self.land_area_change[loss_counts[:,0]] = -plot_size * loss_counts[:,1]
                    elif land_taking_type=='random_percent':
                        # each agent loses a random percentage of their land (until all required land has been taken)
                        order = np.random.choice(self.id, self.N, replace=False)
                        fractions = np.random.uniform(size=self.N)
                        amts = round_up(self.land_area[order] * fractions, self.all_inputs['land']['plot_size'])
                        cum_amts = np.cumsum(amts)
                        cum_losses = np.minimum(cum_amts, area_loss)
                        losses = cum_losses - np.concatenate([[0],cum_losses])[:-1]
                        # assign to agents
                        self.land_area_change[order] = -losses
            else:
                # this is the counterfactual -- assume no land changes
                self.land_area_change = np.full(self.N, 0.)

        self.land_area_post = self.land_area + self.land_area_change
        self.has_land_post = self.land_area_post>0
        np.random.seed(rand_int)
        
        # fix any invalid values
        # NOTE: this doesn't necessarily honor the site-level LUC!! HHs in the counterfactual site might be attributed with less land loss than they experienced
        self.land_area_post[self.land_area_post<0] = 0
        self.land_rented_post[self.land_rented_post<0] = 0
        self.land_rented_post[self.land_rented_post>self.land_area_post] = self.land_area_post[self.land_rented_post>self.land_area_post] # can't rent more land than you have

    def init_farm_size(self):
        '''
        initialize agent-level farm size (ha)
        '''
        mult = self.land_area_multiplier
        if self.N == len(self.land_area_init):
            return np.array(self.land_area_init) * mult
        elif self.N % len(self.land_area_init) == 0:
            # equal number of each
            return np.tile(self.land_area_init, int(self.N / len(self.land_area_init))) * mult
        else:
            return np.random.choice(self.land_area_init, size=self.N) * mult  

    def init_network(self):
        '''
        initialize agent-level networks
        assume each agent's network is the n agents closest to them (i.e., neighbors on each side)
        assuming a 1-d line of agents
        '''
        self.network = np.full([self.N, self.network_n], -1)
        nbr_locs = np.arange(-self.network_n/2, self.network_n/2+1)
        nbr_locs = nbr_locs[nbr_locs!=0]
        for ni in range(self.network_n):
            self.network[:,ni] = np.arange(self.N) + nbr_locs[ni]
        # correct the wrap-around
        self.network[self.network>=self.N] -= self.N

    def income_and_food_consumption(self, market, climate, land, adap_info, estimate=False, objs=False, nf_labor=False, blfs=False, act=False, a=False, prod=False, savings=False):
        '''
        calculate agent income, with subsistence
        if estimate==True, this is in the decision-making and all the other inputs are rqd
        else, use the actual allocations

        keep both of these within one function to ensure they are harmonized
        (note: adap_info refers to different things in the two cases)
        NOTE: the estimate is a bit more stylized than the actual behavior since the estimate is only worrying about net income - not coping measures
        '''
        t = self.t[0]
        if estimate:
            # other income sources and costs throughout the year
            ls_income_tot = objs['ls_obj'][a] * int(self.all_inputs['livestock']['income']) #  + objs['ls_inc'][a] # dimension:(agent)
            salary_income = nf_labor * market.labor_salary # (agent, nZ)
            rqd_costs = self.living_cost * self.living_cost_min_frac # (agent) assume minimum consumption for now.. extra will be added in the coping measures
            rental_costs = self.land_rented * market.land_rental[self.all_inputs['model']['site']]
            outgrower_costs = objs['ha_farmed']['mkt'][a] * (market.farm_cost['mkt'] + 
                market.fertilizer_cost*self.all_inputs['LSLA']['fert_amt_ratio'] * adap_info['fertilizer_application_rate'])            
            # costs at start of year (using savings)
            start_costs = copy.deepcopy(objs['start_costs'][a])
            savings_spent = np.minimum(self.savings[t], start_costs)
            start_costs -= savings_spent
            extra_savings = savings - savings_spent[:,None] if self.savings_acct else 0 # if savings_acct, agents can carry remaining savings over. if not, they can only spend on the start-of-year inputs
            # create an object showing overall cash reserve (including any leftover savings)
            cash = salary_income + extra_savings + (ls_income_tot - start_costs - rqd_costs - rental_costs - outgrower_costs)[:,None] # dimension:(agent,nZ)
            # subsistence + food consumption
            cons_deficit = np.maximum(self.food_rqmt[:,None] - prod['subs'], 0) # (agent, nZ) - mkt crop can't be used for subsistence
            # if not enough production, try to buy the food
            food_costs = cons_deficit * blfs['price_subs'] * market.crop_buy_markup # account for transactions costs
            # sell excess production (subs)
            ag_profits = np.maximum(prod['subs']-self.food_rqmt[:,None], 0) * blfs['price_subs'] # excess subsistence crop goes to mkt
            # sell mkt production.
            if prod['mkt'].sum()>0:
                mkt_prod = copy.deepcopy(prod['mkt']).astype(float)
                # first, format price beliefs if necessary. two things are relevant. (1) outgrower takes away price undertainty, (2) if modeling trust, this affects expected profits
                mkt_price = np.full(blfs['price_mkt'].shape, market.crop_sell['mkt'][t]) if self.all_inputs['adaptation']['outgrower']['fixed_price'] else blfs['price_mkt'] # outgrower takes away mkt uncertainty
                if self.all_inputs['decisions']['outgrower_trust']:
                    mkt_price[blfs['honor_contract']==0] = blfs['price_subs'][blfs['honor_contract']==0] # if modeling trust, set the mkt-price to subs-price for agents that think firm will defect
                    # reduce the actual production if the outgrower breaches
                    mkt_prod[blfs['honor_contract']==0] *= (1-self.all_inputs['adaptation']['outgrower']['breach_frac_prod_lost'])
                # sell
                ag_profits += mkt_prod * mkt_price
            # add to the overall net income figure
            cash += ag_profits - food_costs
            # factor in the value of livestock (otherwise the ls_stock option gets under-valued). don't worry about NPV
            cash += objs['ls_obj'][a][:,None] * market.livestock_cost if self.all_inputs['decisions']['framework']=='util_max' else 0
            # don't worry about food shortages here...   
            return cash
        else:
            # other income sources and costs
            self.ls_income[t] += self.ls_obj * int(self.all_inputs['livestock']['income'])
            self.salary_income[t] = (self.salary_labor[t] * market.labor_salary).astype(int)
            living_cost = (self.living_cost*self.living_cost_min_frac).astype(int) # assume the minimum amount for now...
            rental_costs = self.land_rented * market.land_rental[self.all_inputs['model']['site']]
            outgrower_costs = (land.ha_farmed['mkt'][t] * (market.farm_cost['mkt'] + 
                market.fertilizer_cost*self.all_inputs['LSLA']['fert_amt_ratio'] * self.all_inputs['adaptation']['fertilizer_application_rate'])).astype(int)
            ## SAVINGS + START-OF-YEAR COSTS
            # in all cases, agents can use savings from previous year to pay for these
            start_costs = self.farm_costs[t] + outgrower_costs # farm_costs includes fertilizer
            savings_spent = np.minimum(self.savings[t], start_costs)
            start_costs -= savings_spent # some might be remaining if not enough savings from previous year (though decision-making might have excluded this option)
            # if savings_acct, agents can carry remaining savings over. if not, they can only spend on the start-of-year inputs
            savings = self.savings[t] - savings_spent if self.savings_acct else 0
            
            # create an object showing overall cash reserve
            cash = self.ls_income[t] + self.ls_sell_income[t] + self.salary_income[t] +savings - living_cost - rental_costs - start_costs
            
            ## MARKET PRODUCTION
            # determine the market selling price, based on honoring of contract farming contracts
            prod_mkt = copy.deepcopy(self.crop_production['mkt'][t]).astype(float)
            market_prices = np.full(self.N, market.crop_sell['mkt'][t])
            if prod_mkt.sum() > 0:
                market_prices[~market.honor_contract[t]] = market.crop_sell['subs'][t] # where contract is not honored, agents receive the subsistence price.
                prod_mkt[~market.honor_contract[t]] *= (1-self.all_inputs['adaptation']['outgrower']['breach_frac_prod_lost']) # reduce the overall production for agents that had contract breaches
            # sell market production at this price
            self.farm_income['mkt'][t] += (market_prices*prod_mkt).astype(int)
            cash += self.farm_income['mkt'][t]# add to overall object (NOTE: this does not include the food buying costs)
            ## SUBSISTENCE PRODUCTION
            food_prod = copy.deepcopy(self.crop_production['subs'][t])
            # self.farm_income['subs'][t] += (market.crop_sell['subs'][t] * food_prod).astype(int) # INCLUDE ALL SUBSISTENCE PRODUCTION IN INCOME
            food_amt_sell_debts = np.full(self.N, 0)
            # first, sell any production to absolve debts
            has_debt = cash < 0
            food_sell_debts = np.round(np.minimum(-cash[has_debt], market.crop_sell['subs'][t]*food_prod[has_debt])).astype(int)
            food_amt_sell_debts[has_debt] = np.round(food_sell_debts / market.crop_sell['subs'][t]).astype(int)
            self.food_amt_sell_debts[t] = food_amt_sell_debts
            cash[has_debt] += food_sell_debts
            self.farm_income['subs'][t, has_debt] += food_sell_debts
            food_prod[has_debt] -= food_amt_sell_debts[has_debt]
            # second, sell any excess remaining subsistence food production
            food_amt_sell_excess = np.maximum(food_prod-self.food_rqmt,0).astype(int)
            food_sell_excess = (market.crop_sell['subs'][t] * food_amt_sell_excess).astype(int)
            cash += food_sell_excess
            self.farm_income['subs'][t] += food_sell_excess
            ## FOOD CONSUMPTION
            # need to buy food if there is a deficit in subsistence production (assume buying costs the same as selling)
            cons_deficit = np.maximum(self.food_rqmt-food_prod,0)
            food_cost = (market.crop_sell['subs'][t]*market.crop_buy_markup*cons_deficit).astype(int)
            ## FOOD PURCHASING
            food_bought = np.round(np.maximum(np.minimum(food_cost, cash), 0)).astype(int) # buy the food they can
            self.food_shortage[t] = food_bought < food_cost # FOOD SECURITY
            cash -= food_bought
            food_cost -= food_bought
            self.buy_food[t] = food_bought > 0
            ## DESIRED LIVING COSTS
            # increase expenditure up towards the desired living cost level if possible
            extra_spending = np.maximum(np.minimum(self.living_cost * (1-self.living_cost_min_frac), cash), 0).astype(int)
            cash -= extra_spending
            self.cons_red_rqd[t, cash<=0] = True # ppl that couldn't spend the FULL desired living costs
            # self.savings_post_cons_smooth[t] = copy.deepcopy(self.savings[t+1])

            ## TOTALS
            self.tot_farm_income[t] = self.farm_income['subs'][t] + self.farm_income['mkt'][t]# - outgrower_costs - self.farm_costs[t]
            # create final total income object -- subtract off the value from savings
            self.cash_avail = cash.astype(int)
            self.food_cost_remaining = food_cost
            has_prod = self.tot_crop_production[t] > 0
            self.mkt_orientation[t,has_prod] = (self.crop_production['mkt'][t] + food_amt_sell_debts + food_amt_sell_excess)[has_prod] / self.tot_crop_production[t,has_prod]
            self.cash_expenditure[t] = self.farm_costs[t] + outgrower_costs + living_cost + rental_costs + food_bought + extra_spending
            self.food_expenditure[t] = food_bought
            self.outgrower_costs[t] = outgrower_costs
            self.extra_spending[t] = extra_spending
            self.cash_income[t] = self.ls_income[t] + self.ls_sell_income[t] + self.salary_income[t] + self.tot_farm_income[t]
            self.self_sufficiency[t] = 1 - cons_deficit / self.food_rqmt
            # add to total regional production -- accounting for lost production due to contract breaching
            # land.regional_crop_production[t] += (prod_mkt + self.crop_production['subs'][t]).astype(int).sum()
            # ^^ do this in the land file -- count for all production, even the "lost" stuff

            # save the observations of the contractor's behavior, for those that observe it
            if prod_mkt.sum()>0:
                self.contract_honored[t,land.ha_farmed['mkt'][t]>0] = market.honor_contract[t,land.ha_farmed['mkt'][t]>0].astype(int)

    def coping_measures(self, land, market):
        '''
        simulate coping measures
        assume that the preference for coping is:
        1. casual labor
        2. consumption reduction (food and cash)
        3. livestock sale
        '''
        t = self.t[0]
        food_buy_cost = market.crop_sell['subs'][t] * market.crop_buy_markup
        # calculate the severity of the food shortages
        severity = self.food_cost_remaining / (self.food_rqmt * food_buy_cost)
        coping_food = severity > self.coping_threshold # only agents above the threshold cope. if threshold=0,then all agents cope
        # agents reduce their food consumption to the threshold
        orig_food_cost = copy.deepcopy(self.food_cost_remaining)
        # self.food_cost[t,coping_food] = np.minimum(self.food_cost[t,coping_food], 
        #         (1-self.coping_threshold) * self.food_rqmt[coping_food] * food_buy_cost)
        food_cost_remaining = np.maximum(0, orig_food_cost - self.coping_threshold * self.food_rqmt * food_buy_cost).astype(int) # subtract up to this food amt from the costs
        self.reduce_food_intake[t] = food_cost_remaining < orig_food_cost
        self.reduce_food_intake_amt[t] = (orig_food_cost-food_cost_remaining) / (self.food_rqmt*food_buy_cost) # what fraction of total food rqmt was canceled

        # record who needs to cope
        self.coping_rqd[t] = np.logical_or(self.cash_avail < 0, coping_food) # self.food_cost[t] > 0)
        # calculate how much funds are required
        deficits = np.maximum(-self.cash_avail, 0) + food_cost_remaining

        ## CASUAL LABOR
        if market.wage_labor_coping:
            lbr_rqmt = round_up(np.maximum(deficits/market.labor_wage, 0), market.wage_job_increment) # calculate amt rqd: $ / ($/person) = person
            # code.interact(local=dict(globals(), **locals()))
            max_ppl_avail = round_down(np.maximum(0,self.hh_size-self.salary_labor[t]-self.tot_ag_labor[t]-self.ls_labor[t]), market.wage_job_increment) # should this have -self.salary_labor or not??? yes i think so.
            if (max_ppl_avail<0).sum() > 0:
                print('ERROR: negative labor remaining (agents)')
                code.interact(local=dict(globals(), **locals())) 
            lbr_amts = np.minimum(lbr_rqmt, max_ppl_avail)
            self.wage_labor[t] = market.allocate_wage_labor(self, lbr_amts)
            self.wage_income[t] = self.wage_labor[t] * market.labor_wage
            # attribute to the agents: they buy food first
            food_exp = np.minimum(food_cost_remaining, self.wage_income[t]).astype(int)
            self.cash_avail += (self.wage_income[t] - food_exp) # any extra they keep as cash
            food_cost_remaining -= food_exp
            # save to overall objects
            self.food_expenditure[t] += food_exp
            self.cash_expenditure[t] += food_exp
            self.cash_income[t] += self.wage_income[t]

        ## STRESS DESTOCKING
        # sell livestock if:
        # - food deficit > coping threshold OR
        # - cash deficit remains (i.e., given the minimum cash consumption requirements)
        sell_ls = np.logical_or(food_cost_remaining > 0, self.cash_avail<0)
        deficits = np.maximum(-self.cash_avail, 0) + food_cost_remaining
        sell_rqmt = np.maximum(np.ceil(deficits/market.livestock_cost), 0).astype(int) # calculate amt rqd
        sell_amt = np.minimum(self.ls_obj, sell_rqmt) # restricted by available livestock
        self.ls_obj -= sell_amt # reduce the herdsize
        ls_cash = (sell_amt * market.livestock_cost).astype(int)
        # balance the food and cash with this money
        food_exp = np.minimum(food_cost_remaining, ls_cash).astype(int)
        food_cost_remaining -= food_exp
        self.cash_avail += ls_cash - food_exp # any extra they keep as cash
        # record who has to cope
        self.stress_ls_sell_rqd[t, sell_rqmt>0] = True # record
        self.ls_stress[t] = copy.deepcopy(self.ls_obj)
        # attribute to the overall objects
        self.ls_sell_income[t] += ls_cash
        self.food_expenditure[t] += food_exp
        self.cash_expenditure[t] += food_exp
        self.cash_income[t] += ls_cash

        ## save the savings for next year
        self.savings[t+1] = copy.deepcopy(self.cash_avail)
        # assume that debts can't carry over
        self.cant_cope[t, self.savings[t+1]<0] = True # record
        self.savings[t+1, self.cant_cope[t]] = 0

        # calculate final total labor
        # NOTE: the livestock labor is just based on their estimated residue production and rangeland livestock
        self.total_labor[t] = self.tot_ag_labor[t] + self.ls_labor[t] + self.salary_labor[t] + self.wage_labor[t]
        # if self.total_labor[t].min() == 0:
        
        if np.sum(self.ls_obj<0)>0:
            print('ERROR: negative livestock in agents.coping_measures()')
            code.interact(local=dict(globals(), **locals())) 

    def livestock_stocking_calculation(self, land, rangeland, market):
        '''
        agents calculate the number of livestock they anticipate to be able to keep in the subsequent year,
        given fodder availability.

        if agents have leftover cash and anticipate surplus fodder availability, 
        they will consider purchasing livestock at the start of the subsequent period.
        if there is an anticipated fodder deficit, 
        they will consider destocking at the start of the subsequent period.

        # calculate stocking/destocking
        # stocking 
        # - if the agent has extra money
        # - WITH RANGELAND: if no destocking was rqd at regional level
        # destocking
        # - WITH RANGELAND: already calculated
        # - WITHOUT RANGELAND ONLY: if not enough fodder total
        # if stocking is possible, do not actually stock now. that will be done if the agent decides to
        # in the decision-making at the next time step
        '''
        t = self.t[0]
        ls_inp = self.all_inputs['livestock']
        ## A: how many can be purchased
        self.max_ls_purchase[t] = np.floor(self.savings[t+1] / market.livestock_cost)
        ## B: how many is there labor available for (given current employment and farming)
        # 8/4/2020 -- removing livestock labor restrictions
        # max_labor_tot = np.floor((self.hh_size-self.tot_ag_labor[t]-self.salary_labor[t]) / self.ls_labor_rqmt) # head = ppl / ppl/head
        ## C: how many can be grazed on-farm
        # add the livestock on others' residues
        # average over the past n years
        yrs = np.arange(0,t+1) if t<self.n_yr_smooth else np.arange(t-self.n_yr_smooth+1,t+1)
        max_on_farm = land.residue_production[yrs] / (ls_inp['consumption']) # NOTE: this might be greater than herds_on_residue[t] if the agent had leftover residues...
        max_on_farm += self.herds_on_others_residue[yrs]
        max_on_farm = np.round(np.mean(max_on_farm, axis=0),3)
        ## D: how many can be grazed off-farm
        if t == 0:
            max_off_farm = self.herds_on_rangeland[t]
        else:
            agent_destocks = self.destocking_rqd_rangeland[yrs]
            # take the minimum amt of livestock in years that destocking was required (to be conservative)
            max_off_farm = np.array(np.ma.array(self.herds_on_rangeland[yrs], mask=~agent_destocks).min(axis=0)) # weird issues with fill value being 1e20... but shouldn't matter
            # take the maximum amt of livestock in years that destocking was not required
            max_no_destock = np.array(np.ma.array(self.herds_on_rangeland[yrs], mask=agent_destocks).max(axis=0, fill_value=0))
            # agents that have experienced no destocking assume they're able to increase above the max that they've had by 1
            max_off_farm[max_off_farm>9999] = max_no_destock[max_off_farm>9999] + 1

        # calculate the required change in livestock
        # if this is positive, fodder availability and cash allow for livestock purchase
        # if this is negative (ONLY POSSIBLE W/O RANGELAND) then this represents lack of fodder availability -> destocking
        ls_change = np.min(np.array([self.max_ls_purchase[t], max_on_farm + max_off_farm - self.ls_obj]), axis=0) # max_labor_tot - self.ls_obj
        ls_change = np.floor(ls_change).astype(int)
        # ^^ if this is +ve this represents purchase. 
        # ^^ if it's -ve this represents rqd destocking due to fodder availability (only possible w/o rangeland)

        # de-stock here and only allow for stocking in decision
        # determine where stocking is available and attribute that to the agents
        # they will decide whether to do this at the next time step
        stock = ls_change > 0
        self.ls_stocking_amt[t+1, stock] = ls_change[stock]
        # attribute changes -- destocking only
        # agents destock livestock they don't think they will be able to feed in the next period.
        # this is "compulsory". they can spend the money they make on purchasing inputs etc in the next year.
        # but, if no savings account, they cannot 'save' it throughout the subsequent year.
        # because we don't account for rangeland dynamics + fodder availability in the decision-making,
        # including the choice to destock in decision-making would result in massive livestock over-population
        # because the agents wouldn't anticipate the livestock declines
        self.ls_obj[~stock] += ls_change[~stock] # attribute to livestock
        self.savings[t+1,~stock] -= ls_change[~stock] * market.livestock_cost # attribute to savings
        self.destocking_rqd[t,ls_change<0] = True # record
        self.ls_stocking[t] = copy.deepcopy(self.ls_obj)
        ls_inc = -ls_change[self.destocking_rqd[t]]*market.livestock_cost
        self.ls_sell_income[t, self.destocking_rqd[t]] += ls_inc
        self.cash_income[t, self.destocking_rqd[t]] += ls_inc

        if np.sum(self.ls_obj<0)>0:
            print('ERROR: negative livestock in agents.destocking()')
            code.interact(local=dict(globals(), **locals())) 

def round_up(amts, stepsize):
    '''
    round amts up to the nearest stepsize
    '''
    return np.ceil(amts/stepsize) * stepsize

def round_down(amts, stepsize):
    '''
    round amts up to the nearest stepsize
    '''
    return np.floor(amts/stepsize) * stepsize
