'''
All model input parameters
'''
from collections import OrderedDict
import code

def compile():
    d = {}
    d['model'] = model()
    d['agents'] = agents()
    d['land'] = land()
    d['climate'] = climate()
    d['adaptation'] = adaptation()
    d['rangeland'] = rangeland()
    d['livestock'] = livestock()
    d['LSLA'] = LSLA()
    d['market'] = market()
    d['beliefs'] = beliefs(d['climate'], d['market'])
    d['decisions'] = decisions()
    return d

def model():
    d = {
        'n_agents' : 200,
        'T' : 20, # number of years to simulate (have 20 years of climate data)
        'exp_name' : 'test',
        'seed' : 0,
        'sim_id' : 0,
        'rep_id' : 0,
        'shock' : False,
        # 'lsla_simulation' : False, # does the LSLA happen in this simulation?
        'lsla_experiment' : True, # is this being run as part of an LSLA experiment? If True, then simulate the changes in T and C households
        # ^NOTE: if fitting to empirical data, this should be True because it most accurately represents the conditions in the site
        'matching_experiment' : True, # if True, then read in the HH information from just the HHs that were matched
        'print_init' : False,

        ## if importing agent properties
        'site' : 'OR1',
        'treat' : 'Control', # Treatment or Control
        'LSLA_LUC' : False, # if True, then the LSLA takes farmland and commons
        'matched_ids' : True, # if True, use the empirical hh_ids from the matching experiment. by default, use these.
        'use_weights' : False, # if True (and matched=True) use the household weights rather than the hh-level counterfactuals
        'year_start' : 2000,
        'read_from_file' : ['agents','rangeland','climate'], # which properties to empirically draw from..
        'hh_lulc_data' : 'LULC', # 'LULC' or 'survey' -- use either the HH or the LULC data to model the hh-level land change outcomes.. the HH option includes indirect effects as well by modeling change also in the control sites. the LULC option sets all delta for control site to 0
        'commons_iLUC' : False, # should changes in natural land OUTSIDE the LSLA be considered?
        # other data inputs
        'hh_weights_file' : '../inputs/survey/matching/entropy_weights/{}_fs_days_bin.csv', # the {} should be site
        'hh_match_file' : '../inputs/survey/matching/match_indexes/fs_days_bin.csv',
        'LULC_file' : '../inputs/LULC/site_LULC_2km_buf_for_ABM.csv',
        # 'LULC_file_simplified' : '../inputs/LULC/site_lsla_areas_simplified_2km.csv',
    }
    return d

def adaptation():
    d = {
        'fertilizer' : {
            'application_rate' : 100, # kgN/ha.. site-level means/medians are different. medians PER HA (post) = {or1:100,or2:100,or4:80,or5:100}
        },
        'outgrower' : {
            'join' : False, # can other agents join
            
            ## always need these
            'fixed_price' : True, # given fixed price at start of year?     
            'harvest_time' : '1yr', # 1yr represents maize, 2yr represents sugarcane

            ## only if join==True
            'land_rqmt_type' : 'ha', # {fraction or ha}
            'land_rqmt_amt' : 0.25, # fraction
            'prob_honor' : 1, # probability the firm honors the outgrower agreement. simulate the contract breaches in lsla.py
            'breach_frac_prod_lost' : 0.5, # fraction of production assumed to be lost if the contractor breaches. this could conceptually represent product that is lost on way to market
        },
    }
    d['fertilizer_application_rate'] = d['fertilizer']['application_rate']
    return d

def agents():
    d = {
        # binary switches
        'savings_acct' : False, # if false, agents can't carry over extra money between years. note: 25-50% of HHs have bank accts in post- period. only 10-20% in pre-period..
        'credit_access' : True, # if True, they don't need to be able to afford the option at the start of the year

        #### DEMOGRAPHICS ####
        # data import - if read_from_file==True
        'props_from_file' : ['hh_size','land_area','livestock','decisions'],  # 'land_rental'# reference names for the ABM code
        'file_name' : '../inputs/survey/data_for_analysis.csv',
        # if read_from_file == False:
        'land_area_init' : [0.25,0.5,1, 2], # ha. uniformly sample from each
        'land_area_multiplier' : 1, # for sensitivity analysis
        # living costs
        # 'living_cost_pp' : 1000, # birr/yr. "desired living costs" not including food consumption. 17261 median (hh) value from 2015 LSMS... median non-food expenditure per person in our data: or1(T)=1167, or2T=1475, or4T=2499, or5t=2789
        'living_cost_pp' : {'OR1':1167, 'OR2':1475, 'OR4':2499, 'OR5':2789,'ORX':1983}, # orx is the mean of the others
        'living_cost_min_frac' : 1, # NOW ASSUME THAT THERE IS JUST ONE LEVEL OF EXPENDITURE/LIVING COSTS# 0.5, # fraction of the living costs that must be spent (ie they will sell livestock to do this)
        'food_rqmt_pp' : 18*12, # kg/yr. from resilience paper

        # networks
        'network_n' : 4, # number of neighbors

        # decision-making
        'n_yr_smooth' : 3, # number of smoothing years for livestock management decisions (fodder availability assumption)

        ##### cash + wealth #####
        # initial cash savings (normal distribution)
        'savings_init_mean' : 0, # birr
        'savings_init_sd' : 0,
        'livestock_init' : 2,# 0, (homogeneous for all agents) -- this is overwritten if using data import
        'coping_threshold' : 0.3, # fraction of food deficit required before coping. if 0, then agents cope with any amount of food shortages. if 1 they can reduce their food consumption fully
        
        ##### socio-environmental condns #####
        # NOTE: the model does not restrict agricultural activities to be within the HHs total labor because otherwise the empirical data become infeasible (i.e. there are agents with land+livestock that could otherwise not be feasible in the ABM)
        # BUT the preference for reducing labor still comes in with the satisficing criteria
        # 'labor_constraint_decision' : True,
        'ag_labor_rqmt_subs' : 1, # ppl/ha -- was initially 1.5
        'ag_labor_rqmt_mkt' : 1.5, # ppl/ha -- was initially 1.5
        'ls_labor_rqmt' : 0.2, # ppl/head
    }
    return d

def beliefs(climate, market):
    d = {
        ## beliefs with normal distributions
        'quantities_normal' : ['rain','price_subs','price_mkt'],
        'n0' : [1,1,1], # prior strength on the mean
        'alpha0' : [1,1,1], # prior strength on the variance
        'beta0' : [0.25,0.5,0.5], # E[variance0]
        'mu0' : [climate['rain_mu'],market['crop_sell_params']['subs'][0],market['crop_sell_params']['mkt'][0]], # E[mu0]. the expected value is accurate. this makes the beliefs not affect the results too much

        ## beliefs with binomial distributions
        'network_effects_binom' : True, # influence of observations from agents' networks?
        'quantities_binom' : ['salary_labor_prob','honor_contract'],
        'mu0_binom' : {'salary_labor_prob' : 0.2, 'honor_contract' : 1}, # uninformative belief for contract farming...? but issues when we do small prior strength
        'prior_strength0_binom' : [10, 10], # alpha + beta
    }
    return d

def decisions():
    d = {
        # binary switches
        'framework' : 'util_max', # util_max or imposed or satisficing
        'risk_aversion' : True,
        'outgrower_trust' : False, # does trust in outgrower affect agent decisions?
        'util_tolerance' : 0,#.01, # utility tolerance. when an alternative option is within this tolerance, the household won't change their behavior

        # if framework==imposed
        'imposed_action' : {'salary_labor' : 0, 'fertilizer' : True, 'ls_stock' : True, 'outgrower':0}, # only when t==0 or framework==imposed
        
        # if framework != imposed
        'actions' : OrderedDict({'salary_labor' : [0,-1,1], 'fertilizer' : [False,True], 'ls_stock' : [True,False]}), # it should be such that the first option to be created is the "baseline". numbers for salary_labor are in units of the salary_job_increment parameter
        # ^ NOTE: outgrower:[0,1,-1] is added to this if outgrower['join']=T
        # if risk_aversion = True
        'risk_tolerance_mu' : 3000, # mu for a normal distribution (high = more tolerant to risk)
        'risk_tolerance_cov' : 0.1, # coefficient of variation - i.e., sigma/mu. 0.5 means 50%
        'discount_rate' : 0.586, # fraction. drawn from holden et al 1998 (58.6 = 100/(1+r) --> r = 0.706). this is used for 2-year cash crops
        'nsim_utility' : 10, # number of sims from beliefs when calculating utility
    }
    return d

def market():
    d = {
    # binary switches
    'wage_labor_coping' : True, # if false, exclude wage labor processes as a coping measure

    # 'crop_sell_params' : {'subs' : [2.17,0.25,0.5], 'mkt' : [2.17,0.5,0.5]}, # [x,x,x]=[mean,sd,rho] 2.17 birr/kg is mean 2015 maize price (FAO)
    'crop_sell_params' : {'subs' : [6,0.5,0.7], 'mkt' : [6,0.5,0.7]}, #2.17 [x,x,x]=[mean,sd,rho] 6birr/kg is median maize price in OR sites survey data. 2.17 birr/kg is mean 2015 maize price (FAO). rho is the correlation. should be [0,1]
    'mkt_price_premium' : 1.5, # for the market crop, relative to the subsistence crop
    'crop_buy_markup' : 1.2, # represents transactions costs -- crops are more expensive to buy than to sell. if 1.2 this means 20% more expensive to buy than sell.
    'farm_cost_baseline' : 100, # birr/ha. arbitrary
    'farm_cost_factors' : {'subs' : 1, 'mkt' : 1.5}, # multiplier on baseline. arbitrary
    'fertilizer_cost' : 13.2, # 13.2 birr/kg. median from 2015 LSMS
    'labor_salary' : 18250, # 70*365*5/7, # birr/person/year: 70 birr/day * 5 days per week all year
    'labor_wage' : 18250, # 70*365*5/7, # birr/person/year: 70 birr/day * 5 days per week all year
    'salary_jobs_availability' : 0.1, # full-time salary jobs per agent (used as livelihood strategy)
    'wage_jobs_availability' : 0.1, # single-day wage jobs per agent (used as coping)
    'salary_job_increment' : 0.5, # minimum fraction of a person's time that can be devoted to non-farm SALARY labor (e.g., if 1 then ONLY full-time jobs are available)
    'wage_job_increment' : 0.005, # minimum fraction of a person's time that can be devoted to non-farm WAGE labor (e.g., if 0.005 then 1/200th of year)
    'livestock_cost' : 3000, # birr/head. Ethiopia CSA data 2015
    'land_rental' : {'OR1':4678,'OR2':2266,'OR4':2128,'OR5':1489,'BG3':0,'GM1':0,'GM5':0,'ORX':2640}, # no rental in BG3 or GM1. no costs reported in GM5 so set to zero.. orx is the mean of all OR
    }
    return d

def land():
    d = {
        # binary switches
        'som_dynamics' : False, 
        'labor_yield_reduction' : True,

        'plot_size' : 0.25, # ha. resolution of the landscape. NOTE: not fully integrated into the model
        
        # initial vals
        'organic_N_mu' : 4000, # kgN/ha. similar to initial value in Li2004. if som_dynamics==F, this is the constant value used over all time
        'organic_N_sd' : 200,
        # ag practices
        'fallow_frac' : 0, # fraction fallow under traditional settings. calibrate to get stable SOM
        
        ##### yield #####
        'ag_types' : ['subs','mkt'],
        'max_subs_yield' : 20000, #10k is the tail in the CNH empirical tails 6590,# 6590 kg/ha. maximum, unconstrained yield. 95%ile for Ethiopia-wide LSMS (all 3 years) maize yields
        'max_mkt_yield' : 20000,# 6590,
        'rain_crit' : 0.8, # value at which rainfall starts to be limiting. 0.8 in CENTURY
        'random_effect_sd' : 0.6, # std dev of yield multiplier effect (normal distribution, mu=1)
        'random_effect_mu' : 0,#-0.4,

        ##### residues #####
        'residue_loss_factor' : 0.9, #  90% conversion efficiency  
        'residue_multiplier' : 2, # 2x crop yield->maize residue conversion factor (FAO1987), 
        # 'wealth_N_conversion' : 0.026, # 0.026 kgN/yr per birr. a proxy for livestock manure. derived as 3000birr/head and using values from Newcombe1987. nitrogen %age in manure also similar in Lupwayi2000
        
        ## only relevant if som_dynamics==True
        # 'max_organic_N' : 8000, # REMOVED 2021/1/18 bc hhs with high SOM had unrealistic rainfall factors. now setting this w/in the model. kgN/ha. arbitrary (set in relation to the initial value)
        'fast_mineralization_rate' : 0.5, # 0.6, # what fraction of applied organic matter mineralizes straight away. 0.1 in POM for CC-ins
        'slow_mineralization_rate' : 0.02, # 0.02 rate of mineralization from organic->inorganic (assume linear decay). taken from schmidt2011 -- 50year turnover time of bulk SOM
        'loss_max' : 0.5, # 0.5 inorganic loss fraction with no SOM. Di2002 data had ~50% maximum leaching rates of N. giller1997 says up to 50% in high-rainfall environments
        'loss_min' : 0.05, # 0.05 inorganic loss fraction with maximum SOM. Di2002 had ~5% minimum leaching.
        'fallow_N_add' : 60/0.3, # FOR NOW TRY TO GET NEUTRAL SOIL N BALANCE. this is unrealistic # 40, # kg N/ha. lower limit from N-fixing legumes https://www.tandfonline.com/doi/pdf/10.1080/01904160009382074
        'rain_cropfail_high_SOM' : 0, # rainfall value at which crop yields are 0 with highest SOM. arbitrary
        'rain_cropfail_low_SOM' : 0, #2021/1/18: setting this to the same as high_SOM b/c this paper isn't about SOM changes. now SOM doesn't provide drought benefits.... OLD --> rainfall value at which crop yields are 0 with lowest SOM. arbitrary
        # 'crop_CN_conversion' : 50, # 50 from Century model curves (middle of the y axis) -- pretty arbitrary. represents C:N ratio kind of
        # 'residue_CN_conversion' : 200, # 1/4 of the crop. elias1998

        'crop_N_frac' : 0.012, # kgN/kg dry matter. elias1998 table 8
        'residue_N_frac' : 0.0037, # elias1998 table 8

        # if som_dynamics == False
        # 'rain_cropfail' : 0, # for crop yields
    }
    return d

def climate():
    d = {
        # switches
        'file_name' : '../inputs/climate/precipitation_annual.csv',
        ## ^ NOTE: this just calibrates the variability, not the mean/scale -- this is b/c rainfall differences do not consistently describe site-level yield differences.

        # annual climate measure -- assume normal distribution (truncated to [0,1])
        'rain_mu' : 0.6, # 0.5 approximately fits country-wide CYF distribution for maize (BUT this variable is rain not CYF)
        'rain_sd' : 0.2,

        'shock_years' : [20], # starting at 0 (pythonic)
        'shock_rain' : 0.1, # the rain value in the simulated shock
    }
    return d

def rangeland():
    d = {
        # binary switches
        'integer_consumption' : False, # if true, only an integer number of livestock can be grazed on residues/rangeland
        'reproduce_at_end' : True, # should the reproduction be after the rangeland dynamics. make True as baseline assumption
        'G_constant' : 1500, # grass availability per hectare
    }
    return d

def livestock():
    d = {
        'crop_residue_sharing' : True, # can agents graze livestock on others' leftover crop residues?
        # always used 
        'frac_N_import' : 0, # amount of N from rangeland that can be imported to the farm. controls livestock/SOM feedback
        'N_production' : 78.3, # kg N/year/cattle. see CC_Ins paper for derivation
        'income' : 175, # birr/year/head represents the value of milk production: taken directly from Redda2002 -- 240-480birr/year with local cow. assume 350 and 50% are female --> 175 birr/year/animal
        'consumption' : 2280, # kg/annum (640 in gunnar. i derived 2280 for cc_ins paper: kg dry matter / TLU / year.(Amsalu2014)). also compare to 2700kgDM in (NOTE: ITS AUSTRALIA HERE NOT BORANA) desta1999 for 400kg animal
        'birth_rate' : 0, # use no birth rate -- livestock aren't for breeding. livestock birth rate (probability) (gunnar has 0.8). see also angassa2007 (~0.5)
    }
    return d

def LSLA():
    d = {
        # inputs always rqd
        'type' : 'displacement', # displacement, CF_forced, or CF_join, or none
        
        # 'jobs_per_agent' : {'OR1':0.1875, 'OR2':0.07, 'OR5':0.0185},
        'job_data_source' : 'artificial', # 'empirical' or 'artificial'
        'employment_per_ha' : 0.15, # jobs/ha taken. scheidel2013 has values. 153 jobs / 1000 ha
        'empirical_jobs_per_agent' : {'OR1':0.073, 'OR2':0.03, 'OR5':0},
        # inputs if using data
        'info_file' : '../inputs/lsla_characteristics.csv', # get the start year from this
        'implementation_fraction' : 1,
        'land_taking_type' : 'random_percent', # random_field:take fields randomly or random_percent:sequentially take a random %age of each agent's land
        'fert_amt_ratio' : 1.5, # relative to subsistence rate -- constant over time. set to 0 for no fertilizer or 1 to be the same as subs
        'irrig' : False,

        ## these params will be set based on the "type"
        'LUC' : True, # true for displacement and CF_forced
        'outgrower' : False,

        'frac_smallholder' : 1, # only relevant with ORX experiments
    }
    return d