'''
decision-making functions for the agents
'''
import numpy as np
import pandas as pd
import code
import copy
import sys
from collections import OrderedDict

class Decision():
    def annual_decisions(agents, land, market):
        '''
        wrapper function for decision-making
        '''
        # different behavior in 1st year
        inp = agents.all_inputs['decisions']
        agents.ls_obj = agents.livestock[agents.t[0]] # initialize a temporary object to be worked with
        agents.ls_start[agents.t[0]] = copy.deepcopy(agents.ls_obj)

        # assume no fallow for now
        agents.fallow[agents.t[0]] = False

        #### FOR CALIBRATION: INPUT BASELINE DECISIONS AT T=0??
        if ((agents.t[0]==0) * (agents.file_init * ('decisions' in agents.props_from_file))):
            Decision.empirical_decisions(agents, land, market, inp) # initialize from the empirical data
        elif inp['framework'] == 'imposed':
            Decision.imposed_decision(agents, land, market, inp) # no decisions
        elif inp['framework'] == 'util_max':
            Decision.utility_max(agents, land, market, inp)
        elif inp['framework'] == 'satisficing':
            Decision.satisficing(agents, land, market, inp)
        # print(agents.choices['ls_stock'][agents.t[0]].mean())

    def empirical_decisions(agents, land, market, inp):
        '''
        impose a single behavior for all agents
        '''
        t = agents.t[0]
        act = inp['imposed_action']
        # init object to save livelihood allocations
        objs = Decision.init_decision_objects(agents, len(agents.decision_options), inp['nsim_utility']) # single option
        # init beliefs (not really necessary for this option)
        blfs = Decision.format_beliefs(agents, t) # dict with each element of size [agent,nZ]
        # evaluate the options
        choices = np.full(agents.N, -99)
        for a, act in enumerate(agents.decision_options):
            # find the indices of the agents that choose this option
            ixa = np.full(agents.N, True)
            for act_a, val_a in act.items():
                ixa *= (agents.decisions_init[act_a] == val_a)
            choices[ixa] = a
            # evaluate it if there are some doing this..
            if ixa.sum()>0:
                [objs, xx, xyz] = Decision.evaluate_option(a, act, objs, agents, land, market, inp, blfs)
        
        agents.choice_ixs[t] = choices
        # attribute the outcomes
        Decision.attribute_decisions(agents, choices, land, market, objs, t)    

    def imposed_decision(agents, land, market, inp):
        '''
        impose a single behavior for all agents
        '''
        t = agents.t[0]
        act = inp['imposed_action']
        # init object to save livelihood allocations
        objs = Decision.init_decision_objects(agents, len(agents.decision_options), inp['nsim_utility']) # single option
        # init beliefs (not really necessary for this option)
        blfs = Decision.format_beliefs(agents, t) # dict with each element of size [agent,nZ]
        # select the "choice" -- it's the imposed one..
        for i, opt in enumerate(agents.decision_options):
            fine = True
            for k, v in opt.items():
                fine *= True if act[k] == v else False
            if fine:
                choice_ix = i
        
        # evaluate the option
        [objs, xx, xyz] = Decision.evaluate_option(choice_ix, agents.decision_options[choice_ix], objs, agents, land, market, inp, blfs)
        choices = np.full(agents.N, choice_ix)
        agents.choice_ixs[t] = choices
        # attribute the outcomes
        Decision.attribute_decisions(agents, choices, land, market, objs, t)        
    
    def satisficing(agents, land, market, inp):
        '''
        make decisions using satisficing framework:
        preference 1: non-negative net income
        preference 2: leisure time
        --> choose the option that has the lowest probability of negative income
        --> if multiple tie, choose the one with most leisure time
        '''
        t = agents.t[0]
        # format beliefs for utility calculations
        blfs = Decision.format_beliefs(agents, t) # dict with each element of size [agent,nZ]
        # initialize the objects to store the decision outcomes
        objs = Decision.init_decision_objects(agents, len(agents.decision_options), inp['nsim_utility'])
        exp_util = np.full([len(agents.decision_options), agents.N], np.nan)
        utils = np.full([len(agents.decision_options), agents.N, inp['nsim_utility']], np.nan)

        # loop over the options
        for a, act in enumerate(agents.decision_options):
            if ('outgrower' in act.keys()) and (act['outgrower'] != 0) and (not agents.can_join_outgrower):
                # agents can't join/leave outgrower, so don't waste time on this
                exp_util[a] = -99
                utils[a] = -99
                agents.option_feasibility[t,a] = False
            else:
                # evaluate option normally
                [objs, exp_util[a], utils[a]] = Decision.evaluate_option(a, act, objs, agents, land, market, inp, blfs)
            
        # evaluate the probability of positive income
        # note: the mapping of (+/-) in income to (+/-) in utility is consistent
        agents.exp_util[t] = copy.deepcopy(exp_util)
        agents.exp_util[t][~agents.option_feasibility[t]] = -99999
        p_pos = np.mean(utils>0, axis=2)
        # assume that infeasible options cannot have positive income so shouldn't be accepted
        # the 0th option is the baseline, which we assume is feasible
        p_pos[1:][~agents.option_feasibility[t,1:]] = -1
        # find maximum probability
        max_p = np.max(p_pos, axis=0)
        max_p_locs = p_pos==max_p
        n_max_p = np.sum(max_p_locs, axis=0)
        agents.pos_util[t] = p_pos

        ### for agents with a single option: choose this
        choices = np.full(agents.N, -99)
        choices[n_max_p==1] = np.argmax(p_pos, axis=0)[n_max_p==1]
        
        ### for all other agents: maximize leisure -- use the expected amount of non-farm labor
        leis = n_max_p>1
        tot_work = (objs['nf_labor'].mean(axis=2) + objs['tot_ag_lbr'] + objs['ls_labor'])[:,leis]
        tot_work[~np.transpose(agents.option_feasibility[t,:,leis])] = 99 # infeasible options are burdensome
        tot_work[~max_p_locs[:,leis]] = 99 # also restrict to the maximum probability options
        tot_work = np.round(tot_work,2)
        min_work = np.min(tot_work, axis=0)

        # select the min_loc. 
        # if multiple options are very close, then choose what they've done most recently from these
        # if multiple options are very close but none have been done, choose randomly
        rand_int = np.random.randint(1e6) # generate random integer to control stochasticity
        tols = np.abs(tot_work-min_work[None,:]) <= inp['util_tolerance'] # closeness threshold
        # identify agents whose previous decision is within the tolerance
        prev_choices = agents.choice_ixs[t-1,leis]
        prev_in_tols = tols[prev_choices, np.arange(leis.sum())]
        ids = np.arange(agents.N)[leis][prev_in_tols]
        choices[ids] = agents.choice_ixs[t-1,ids]
        # if not, choose randomly from within the next most similar remaining decision (i.e. one of the activities might match)
        # vectorized form. from https://stackoverflow.com/questions/47722005/vectorizing-numpy-random-choice-for-given-2d-array-of-probabilities-along-an-a
        max_similarity = agents.decision_similarity.max()
        sim_vals = np.transpose((agents.decision_similarity==(max_similarity-1))[prev_choices])
        similar_in_tol = sim_vals[:,~prev_in_tols] * tols[:,~prev_in_tols]
        has_similar_in_tol = similar_in_tol.sum(0)>0
        similar_in_tol_valid = similar_in_tol[:,has_similar_in_tol]
        rndm_choices = ((similar_in_tol_valid.cumsum(0) / similar_in_tol_valid.sum(0)) > np.random.rand(similar_in_tol_valid.shape[1])[None,:]).argmax(0)
        ids = np.arange(agents.N)[leis][~prev_in_tols][has_similar_in_tol]
        choices[ids] = rndm_choices
        # finally, choose randomly from within tols if nothing is similar
        no_similar = tols[:,~prev_in_tols][:,~has_similar_in_tol]
        rndm_choices = ((no_similar.cumsum(0) / no_similar.sum(0)) > np.random.rand(no_similar.shape[1])[None,:]).argmax(0)
        ids = np.arange(agents.N)[leis][~prev_in_tols][~has_similar_in_tol]
        choices[ids] = rndm_choices
        np.random.seed(rand_int)

        ## SPECIAL CASE:
        # for agents with no good option (i.e., max_p==0), choose the option with the maximum utility
        all_bad = max_p==0
        max_utils = agents.exp_util[t,:,all_bad].max(1)
        nmax = (agents.exp_util[t,:,all_bad]==max_utils[:,None]).sum(1)
        # choices[all_bad] = sdfa
        # NEED TO FIGURE OUT WHAT TO DO!! AGENTS HAVE MULTIPLE OPTIONS THAT TIE
        # ALSO, UTILITY IS THE SAME FOR FERT AND NO_FERT FOR AGENTS WITH NO LAND .... WHAT TO THEY ACTUALLY CHOOSE?

        # attribute the outcomes
        agents.choice_ixs[t] = choices
        Decision.attribute_decisions(agents, choices, land, market, objs, t)
   
    def utility_max(agents, land, market, inp):
        '''
        make decisions using utility maximization framework
        '''
        t = agents.t[0]
        # format beliefs for utility calculations
        blfs = Decision.format_beliefs(agents, t) # dict with each element of size [agent,nZ]
        # initialize the objects to store the decision outcomes
        objs = Decision.init_decision_objects(agents, len(agents.decision_options), inp['nsim_utility'])
        exp_util = np.full([len(agents.decision_options), agents.N], np.nan)
        utils = np.full([len(agents.decision_options), agents.N, inp['nsim_utility']], np.nan)

        # loop over the options
        for a, act in enumerate(agents.decision_options):
            if ('outgrower' in act.keys()) and (act['outgrower'] != 0) and (not agents.can_join_outgrower):
                # agents can't join/leave outgrower, so don't waste time on this
                exp_util[a] = -99
                utils[a] = -99
                agents.option_feasibility[t,a] = False
            else:
                # evaluate option normally
                [objs, exp_util[a], utils[a]] = Decision.evaluate_option(a, act, objs, agents, land, market, inp, blfs)

        # select the option for each agent with the highest utility
        agents.exp_util[t] = copy.deepcopy(exp_util)
        ### effects of feasbility
        exp_util[~agents.option_feasibility[t]] = -99999
        # BUT assume that the first option is the "baseline", so is always feasible
        exp_util[0] = agents.exp_util[t,0]
        # select the best feasible option
        # if multiple options are very close, then choose what they've done most recently from these
        # if multiple options are very close but none have been done, choose randomly
        rand_int = np.random.randint(1e6) # generate random integer to control stochasticity
        choices = np.full(agents.N, -99)
        maxs = exp_util.max(0) # identify maximum utility
        agents.max_utility[t] = maxs
        tols = np.abs(exp_util-maxs[None,:]) <= inp['util_tolerance'] # closeness threshold
        # identify agents whose previous decision is within the tolerance
        prev_choices = agents.choice_ixs[t-1]
        prev_in_tols = tols[prev_choices, np.arange(agents.N)]
        choices[prev_in_tols] = agents.choice_ixs[t-1,prev_in_tols]
        # if not, choose randomly from within the next most similar remaining decision (i.e. one of the activities might match)
        # vectorized form. from https://stackoverflow.com/questions/47722005/vectorizing-numpy-random-choice-for-given-2d-array-of-probabilities-along-an-a
        max_similarity = agents.decision_similarity.max()
        sim_vals = np.transpose((agents.decision_similarity==(max_similarity-1))[prev_choices])
        similar_in_tol = sim_vals[:,~prev_in_tols] * tols[:,~prev_in_tols]
        has_similar_in_tol = similar_in_tol.sum(0)>0
        similar_in_tol_valid = similar_in_tol[:,has_similar_in_tol]
        rndm_choices = ((similar_in_tol_valid.cumsum(0) / similar_in_tol_valid.sum(0)) > np.random.rand(similar_in_tol_valid.shape[1])[None,:]).argmax(0)
        ids = agents.id[~prev_in_tols][has_similar_in_tol]
        choices[ids] = rndm_choices
        # finally, choose randomly from within tols if nothing is similar
        no_similar = tols[:,~prev_in_tols][:,~has_similar_in_tol]
        rndm_choices = ((no_similar.cumsum(0) / no_similar.sum(0)) > np.random.rand(no_similar.shape[1])[None,:]).argmax(0)
        ids = agents.id[~prev_in_tols][~has_similar_in_tol]
        choices[ids] = rndm_choices
        np.random.seed(rand_int)

        # attribute the outcomes
        agents.choice_ixs[t] = choices
        Decision.attribute_decisions(agents, choices, land, market, objs, t)

    def init_decision_objects(agents, n_decisions, nZ):
        # init some objects to save temporary information
        # at the end, draw from this to assign agents' chosen options
        na = n_decisions
        objs = {
            'tot_ha_farmed' : np.full([na, agents.N], np.nan),
            'ha_farmed' : {'mkt' : np.full([na, agents.N], np.nan), 'subs' : np.full([na, agents.N], np.nan)},
            'start_costs' : np.full([na, agents.N], 0),
            'ag_lbr' : {'mkt' : np.full([na, agents.N], np.nan), 'subs' : np.full([na, agents.N], np.nan)},
            'tot_ag_lbr' : np.full([na, agents.N], np.nan),
            'lbr_tot' : np.full([na, agents.N], np.nan),
            'ls_labor' : np.full([na, agents.N], np.nan),
            'ls_obj' : np.full([na, agents.N], 0),
            'nf_labor' : np.full([na, agents.N, nZ], np.nan), # uncertainty
            'prod' : {'mkt' : np.full([na, agents.N, nZ], np.nan), 'subs' : np.full([na, agents.N, nZ], np.nan)}, # uncertainty
        }
        return objs

    def evaluate_option(a, act, objs, agents, land, market, inp, blfs):
        # init some variables
        t = agents.t[0]
        adap_info = agents.all_inputs['adaptation']
        nZ = inp['nsim_utility']
        outg_info = agents.all_inputs['adaptation']['outgrower']
        risk_tolerances = np.repeat(agents.risk_tolerance, nZ).reshape(agents.N, nZ)

        ##### ALLOCATE LAND #####
        [tot_ha_farmed, ha_farmed, crops] = Decision.allocate_land(adap_info, act, agents, land, outg_info, t)

        ##### ALLOCATE CASH #####
        ## calculate the start-of-year costs of the option (this is the same each year)
        # note: this assumes that agents only grow mkt crops with the outgrower, which gives inputs on loan, so agents never have start-of-year costs for this crop type
        start_costs = market.farm_cost['subs'] * ha_farmed['subs']
        if 'fertilizer' in act.keys() and act['fertilizer']:
            start_costs += market.fertilizer_cost*ha_farmed['subs']*adap_info['fertilizer_application_rate'] # birr/kg * ha_farmed * kg/ha = birr

        ##### LIVESTOCK STOCKING #####
        ls_obj = copy.deepcopy(agents.ls_obj) # create new livestock object
        if 'ls_stock' in act.keys() and act['ls_stock']:
            amts = agents.ls_stocking_amt[t]
            ls_obj += amts
            start_costs += amts * market.livestock_cost# * market.crop_buy_markup # include the mark-up in livestock also??

        ##### ALLOCATE LABOR #####
        [ag_lbr, tot_ag_lbr, ls_labor, nf_labor, nf_labor_max, lbr_tot, ha_farmed] = Decision.allocate_labor(agents, land, ha_farmed, t, nZ, act, market, crops, blfs, ls_obj)

        ##### FEASIBILITY #####
        agents.option_feasibility[t,a] = (lbr_tot >= 0) * (nf_labor.min(1) >= 0) * (nf_labor_max <= agents.hh_size) # labor rqmts
        agents.option_feasibility[t,a] *= (nf_labor_max <= np.maximum(0, agents.hh_size-tot_ag_lbr-ls_labor)) # non-farm labor rqmts
        agents.option_feasibility[t,a] *= (ha_farmed['mkt'] <= tot_ha_farmed) * (ha_farmed['mkt'] >= 0) # land rqmts
        if not agents.credit_access:
            agents.option_feasibility[t,a] *= (start_costs <= agents.savings[t]) 
        # outgrower restrictions
        if agents.forced_outgrower.sum()>0:
            # agents forced to be in outgrower cannot decrease their participation below the forced amt of land
            invalid = agents.forced_outgrower * (ha_farmed['mkt'] < agents.forced_outgrower_ha)
            agents.option_feasibility[t,a,invalid] = False
        if ('outgrower' in act.keys()):
            if (act['outgrower'] != 0) and (not agents.can_join_outgrower):
                # can't do outgrower if the LSLA isn't active yet -- this should have already been filtered in satisficing/util_max, but do it here also
                agents.option_feasibility[t,a] = False
            if (act['outgrower'] == -1) and (agents.can_join_outgrower) and (outg_info['harvest_time']=='2yr'):
                # can't abandon sugarcane half way through
                invalid = ha_farmed['mkt'] < land.ha_mkt_yr1[t-1]
                agents.option_feasibility[t,a,invalid] = False
        # feasibility of livestock stocking -- infeasible if no stocking ability
        if 'ls_stock' in act and act['ls_stock']:
            agents.option_feasibility[t,a,agents.ls_stocking_amt[t]==0] = False
        # agents with no land can't do fertilizer
        if 'fertilizer' in act and act['fertilizer']:
            agents.option_feasibility[t,a,tot_ha_farmed==0] = False
        
        ## estimate crop production, given beliefs
        prod = Decision.estimate_crop_yields(agents, land, blfs, adap_info, act, t, nZ, ha_farmed, lbr_tot) 

        # save the allocations
        # note: the variable name needs to be the same as the object in memory
        for k, v in objs.items():
            if isinstance(eval(k), dict):
                for k2, v2 in eval(k).items():
                    v[k2][a] = v2
            else:
                v[a] = eval(k)

        # calculate net income, incorporating uncertainty in climate, prices, and outgrower contract breaching
        # accounting for subsistence requirements
        savings_nZ = np.repeat(agents.savings[t], nZ).reshape(agents.N, nZ)
        if ((ha_farmed['mkt'] != 0).sum() == 0) or (outg_info['harvest_time'] == '1yr'):
            # single-year calculation (if nobody doing mkt or single-year)
            net_income = agents.income_and_food_consumption(market, False, land, adap_info, True, objs, nf_labor, blfs, act, a, prod, savings_nZ)
        else:
            # market crops are only harvested at the end of the 2nd year..
            ## year 1: market production from the crops that are ready to be harvested
            ha_harvest = land.ha_mkt_yr1[t-1] # crops that were in their first year last year
            valid = (ha_harvest > 0) * (ha_farmed['mkt'] > 0)
            mkt_prod1 = np.full(prod['mkt'].shape, 0)
            mkt_prod1[valid] = 2*prod['mkt'][valid] * (ha_harvest[valid]/ha_farmed['mkt'][valid])[:,None] # multiply production by two b/c two years of growth (proxies the growth they already have)
            prod1 = {'subs':prod['subs'], 'mkt':mkt_prod1}
            cash1 = agents.income_and_food_consumption(market, False, land, adap_info, True, objs, nf_labor, blfs, act, a, prod1, savings_nZ)
            ## year 2: realize market production and carry over the cash from the first year.
            cash1_reset = copy.deepcopy(cash1)
            cash1_reset[cash1_reset<0] = 0 # can't carry over negative savings
            prod2 = {'subs':prod['subs'], 'mkt':2*prod['mkt'] - mkt_prod1} # assume cumulative production over the two years, subtracting the harvests from the first year
            cash2 = agents.income_and_food_consumption(market, False, land, adap_info, True, objs, nf_labor, blfs, act, a, prod2, cash1_reset)
            # combine
            npv_vals = np.array([agents.discount_rate**i for i in range(2)])
            net_income = np.array([cash1,cash2]) * npv_vals[:,None,None]
            risk_tolerances = np.array([risk_tolerances,risk_tolerances])
            # code.interact(local=dict(globals(), **locals()))
        
        # convert to utility (dimension: (agent,nZ))
        if agents.all_inputs['decisions']['risk_aversion']:
            pos = net_income > 0   
            rndm_utils = np.full(net_income.shape, np.nan)
            rndm_utils[pos] = 1 - np.exp(-net_income[pos] / risk_tolerances[pos])
            rndm_utils[~pos] = -(1 - np.exp(net_income[~pos] / risk_tolerances[~pos])) # assume risk averion = loss aversion (note: this is NOT empirically justified in decision theory)
            # import matplotlib.pyplot as plt
            # fig, ax = plt.subplots()
            # ax.scatter(net_income.flatten(),rndm_utils.flatten())
            # fig.savefig('utils_{}.png'.format(agents.all_inputs['decisions']['risk_tolerance_params'][0]))
            # sys.exit()
        else:
            rndm_utils = net_income
        
        # 2e: calculate expected utility
        if ((ha_farmed['mkt'] != 0).sum() > 0) and (outg_info['harvest_time'] == '2yr'):
            rndm_utils[1,ha_farmed['mkt']==0] = rndm_utils[0,ha_farmed['mkt']==0] # assume agents with no 2-year crop just do a single-horizon calculation
            exp_util = np.mean(rndm_utils, axis=(0,2)) # dimension : (agent) -- calculate average over the two years
            rndm_utils = rndm_utils[0] # first year only. this object is only used with satisficing decision-making. assume satisficers are conservative and don't want negative income/utility in the first year
        else:
            exp_util = np.mean(rndm_utils, axis=1) # dimension : (agent)

        return objs, exp_util, rndm_utils
    
    def format_beliefs(agents, t):
        mu_np = np.array([agents.blf.mu[i][t] for i in agents.blf.quantities_normal]) # dimension: [blf_type, agent]
        sd_np = np.sqrt(np.array([agents.blf.var[i][t] for i in agents.blf.quantities_normal])) # [blf_type, agent]
        # integrate with the random normal simulations
        # this assumes that the uncertainty in all beliefs is perfectly correlated
        # e.g. a bad rainfall year is also a bad price year
        blf_dist = agents.rndm_Zs[t][None,:,:] * sd_np[:,:,None] + mu_np[:,:,None] # dimension: (agent,nZ) * (blf_type, agent) --> (blf_type,agent,nZ)
        blfs = {}
        for qi, quant in enumerate(agents.blf.quantities_normal):
            blfs[quant] = blf_dist[qi] # shape: (agent,nZ)
            # constrain
            blfs[quant][blfs[quant]<0] = 0
            if quant == 'rain':
                blfs[quant][blfs[quant]>1] = 1
        
        # now the binomial quantities
        for qi, quant in enumerate(agents.blf.quantities_binom):
            blfs[quant] = (agents.rndm_Us[t] <= agents.blf.mu[quant][t][:,None]).astype(int)

        return blfs

    def allocate_land(adap_info, act, agents, land, outg_info, t):
        '''
        note: for outgrower, this does not distinguish between the year of the crop for 2-year crops
        '''
        cons_fac = (1-adap_info['conservation']['area_req']*act['conservation']) if 'conservation' in act.keys() else 1
        tot_ha_farmed = agents.land_area * cons_fac * (1-land.fallow_frac * agents.fallow[t])
        ha_farmed = {}
        if 'outgrower' in act.keys():
            prev_mkt = land.ha_farmed['mkt'][t-1] if t > 0 else np.full(agents.N, 0.)
            ha_farmed['mkt'] = prev_mkt + act['outgrower'] * outg_info['land_rqmt_amt'] # 0=no change, 1=increase, -1=decrease
        else:
            # assume market-based crops only grown in outgrower
            ha_farmed['mkt'] = np.full(agents.N, 0.)
        # all remaining area is subsistence
        ha_farmed['subs'] = np.maximum(tot_ha_farmed - ha_farmed['mkt'],0) # in case mkt > tot_ha_farmed
        
        # modify the values for the HHs forced into the outgrower scheme
        # assume for now that all of their land is in the outgrower scheme
        if agents.forced_outgrower.sum()>0:
            ha_farmed['mkt'][agents.forced_outgrower] = np.minimum(agents.forced_outgrower_ha, tot_ha_farmed)[agents.forced_outgrower]
            ha_farmed['subs'][agents.forced_outgrower] = (tot_ha_farmed-ha_farmed['mkt'])[agents.forced_outgrower]

        crops = ['subs','mkt'] if max(ha_farmed['mkt'])>0 else ['subs'] # don't bother with mkt if nobody is growing it (for computational eff.)

        return tot_ha_farmed, ha_farmed, crops

    def allocate_labor(agents, land, ha_farmed, t, nZ, act, market, crops, blfs, ls_obj):
        '''
        NOTE 8/4/2020:
        when parameterizing using the empirical data, many HHs become infeasible based on labor requirements
        (ie their initial land cultivation + livestock holdings are such that their labor is more than hh_size)
        SO i am removing restrictions based on labor requirements
        there is still a preference for reducing labor in the satisficing criteria, but not a restriction per se
        '''
        # farm labor
        ag_lbr = {} # init temp object
        tot_ag_lbr = np.full(agents.N, 0.) # init temp object
        for crop in crops:
            ag_lbr[crop] = agents.ag_labor_rqmt[crop]*ha_farmed[crop] # no restrictions on ag_lbr (even if it is larger than hh_size)
            tot_ag_lbr += ag_lbr[crop] # add to total
       
        ## non-farm salary labor
        # note1: there is inertia in this from last time
        # note2: there is uncertainty in the ability to get a _new_ job (i.e. when increasing nf_labor)
        #        we model the uncertainty in this using the agents' beliefs about the probability they will get the job
        # start by creating object of size [agent,nZ] to allow for this
        nf_labor = np.repeat(agents.salary_labor[t-1], nZ).reshape(agents.N, nZ) # last year's values
        nf_labor_max = copy.deepcopy(nf_labor).mean(1) # making this obj, b/c some people might believe they're not going to get a job but then they do and this messes up their labor allocation...
        if 'salary_labor' in act.keys():
            if act['salary_labor'] == -1:
                # decrease
                nf_labor -= market.salary_job_increment
                nf_labor_max -= market.salary_job_increment
            elif act['salary_labor'] == 1:
                # increase -- here there is uncertainty in the blfs object
                nf_labor += blfs['salary_labor_prob'] * market.salary_job_increment
                nf_labor_max += market.salary_job_increment # most optimistic view
        
        ## livestock labor
        # assume no destocking based on labor requirements
        # and labor requirements only for livestock that are off-farm
        # first, calculate max livestock on-farm
        if t == 0:
            # no information -- assume we gucci
            max_on_farm = copy.deepcopy(ls_obj)
        else:
            yrs = np.arange(0,t) if (t-1)<agents.n_yr_smooth else np.arange(t-agents.n_yr_smooth,t) # average over the past n years
            max_on_farm = land.residue_production[yrs] / (agents.all_inputs['livestock']['consumption']) # NOTE: this might be greater than herds_on_residue[t] if the agent had leftover residues...
            max_on_farm += agents.herds_on_others_residue[yrs]
            max_on_farm = np.round(np.mean(max_on_farm, axis=0),3)

        # now calculate the labor rqmt for the livestock off-farm
        ls_off_farm = np.maximum(ls_obj - max_on_farm, 0)
        ls_labor = ls_off_farm * agents.ls_labor_rqmt

        lbr_tot = tot_ag_lbr + ls_labor + nf_labor_max # nf_labor.max(axis=1) # assume the most optimistic w.r.t. non-farm jobs

        if (ls_obj < 0).sum()>0:
            print('ERROR: negative livestock remaining (decision)')
            code.interact(local=dict(globals(), **locals()))
        
        return ag_lbr, tot_ag_lbr, ls_labor, nf_labor, nf_labor_max, lbr_tot, ha_farmed
    
    def estimate_crop_yields(agents, land, blfs, adap_info, act, t, nZ, ha_farmed, lbr_tot):
        prod = {}
        for crop in land.ag_types:
            # water constraints
            y_w = blfs['rain'] * land.max_yield[crop] # shape: (agent, nZ)
            # nutrient availability
            N_available = land.slow_mineralization_rate * land.organic[t] # assume perfect information about current SOM
            if crop == 'subs':
                if 'fertilizer' in act.keys():
                    N_available += adap_info['fertilizer_application_rate'] * act['fertilizer']
            elif crop == 'mkt':
                N_available += agents.all_inputs['LSLA']['fert_amt_ratio'] * adap_info['fertilizer_application_rate'] # only ever grow market crop with outgrower

            # calculate the max yield, given available nutrients. assume that some growth goes to residues 
            y_n = N_available / (land.crop_N_frac+land.residue_multiplier*land.residue_N_frac) #  shape: (agent). kgN/ha / (kgN/kgC_yield) = kgC/ha ~= yield(perha)
            # combine
            crop_yield = np.minimum(y_w, y_n[:,None]).astype(int) # shape: (agent,nZ)
            # add labor yield effects, if appropriate
            if land.labor_yield_reduction:
                lbr_eff = 1 - np.exp(-agents.hh_size[lbr_tot>0] / lbr_tot[lbr_tot>0])
                crop_yield[lbr_tot>0] = (crop_yield[lbr_tot>0] * lbr_eff[:,None]).astype(int)
            # total production
            prod[crop] = (crop_yield * ha_farmed[crop][:,None]).astype(int) # dimension: (agent, nZ)

        return prod

    def attribute_decisions(agents, choices, land, market, objs, t):
        '''
        attribute the agents' choices to the agents and land objects
        '''   
        ixx = [choices, np.arange(agents.N)] # combined index of the selected option for each agent
        
        # make the allocations
        for crop in land.ag_types:
            land.ha_farmed[crop][t] = objs['ha_farmed'][crop][tuple(ixx)]
            agents.ag_labor[crop][t] = objs['ag_lbr'][crop][tuple(ixx)]

        for a, act in enumerate(agents.decision_options):
            ixa = choices == a
            if 'fertilizer' in act.keys() and act['fertilizer']:
                land.fertilizer['subs'][t,ixa] = agents.all_inputs['adaptation']['fertilizer_application_rate'] # kg N/ha farmed
            
            for k, v in act.items():
                agents.choices[k][t,ixa] = v

        agents.farm_costs[t] = objs['start_costs'][tuple(ixx)] # includes livestock stocking
        agents.tot_ag_labor[t] = objs['tot_ag_lbr'][tuple(ixx)]
        agents.ls_obj = objs['ls_obj'][tuple(ixx)]
        agents.ls_decision[t] = copy.deepcopy(agents.ls_obj)
        agents.ls_labor[t] = objs['ls_labor'][tuple(ixx)]
        agents.salary_consider[t,agents.choices['salary_labor'][t]==1] = True
        agents.nf_labor_temp = objs['nf_labor']
        # outgrower: those forced and those who chose it
        agents.outgrower_yr[t] = (land.ha_farmed['mkt'][t] > 0)
        # 2-year crops: attribute to land (note which is 2nd year and which is 1st year)
        if (agents.all_inputs['adaptation']['outgrower']['harvest_time'] == '2yr') and land.ha_farmed['mkt'][t].sum()>0:
            land.ha_mkt_yr1[t] = land.ha_farmed['mkt'][t] - land.ha_mkt_yr1[t-1]
            land.ha_mkt_yr2[t] = land.ha_farmed['mkt'][t] - land.ha_mkt_yr1[t]

def round_up(amts, stepsize):
    '''
    round amts up to the nearest stepsize
    '''
    return np.ceil(amts/stepsize) * stepsize

def round_down(amts, stepsize):
    '''
    round amts up to the nearest stepsize
    '''
    return np.floor(amts/stepsize) * stepsize