import numpy as np
import code
import copy

class Land():
    def __init__(self, agents, inputs):
        # attribute the parameters to the object
        self.all_inputs = inputs
        self.inputs = inputs['land']
        for key, val in self.inputs.items():
            setattr(self, key, val)
        self.T = self.all_inputs['model']['T']

        # how many total plots?
        self.n_plots = agents.N
        self.tot_area = np.sum(agents.land_area)
        self.owner = agents.id

        ##### soil properties #####
        # assume homogeneous soil properties over each agent's land
        # i.e., if they do multiple types of ag, assume crops are rotated
        # represents the START of the year
        self.organic = np.full([self.T+1, self.n_plots], np.nan)
        try:
            som_orx = np.mean([self.organic_N_mu_OR1,self.organic_N_mu_OR2,self.organic_N_mu_OR4,self.organic_N_mu_OR5])
            som_values = {'OR1':self.organic_N_mu_OR1,'OR2':self.organic_N_mu_OR2,'OR4':self.organic_N_mu_OR4,'OR5':self.organic_N_mu_OR5,'ORX':som_orx}
            som_value_site = som_values[self.all_inputs['model']['site']]
        except:
            som_value_site = self.organic_N_mu
        self.organic[0] = np.random.lognormal(som_value_site, self.organic_N_sd, size=self.n_plots) # kg N/ha
        # code.interact(local=dict(globals(), **locals()))
        # self.organic[0] = np.random.normal(self.organic_N_mu, self.organic_N_sd, size=self.n_plots) # kg N/ha
        self.organic[0][self.organic[0]<0] = 0
        self.max_organic_N = self.organic[0].max() # set as max within the population, otherwise some ppl have rainfall effects that don't make sense
        # self.organic[0][self.organic[0]>self.max_organic_N] = self.max_organic_N
        # inorganic represents the END of the year (i.e. for crop yield)
        # there can be separate amts for different crop mgmt types
        self.inorganic = {'subs' : np.full([self.T, self.n_plots], np.nan), 'mkt' : np.full([self.T, self.n_plots], np.nan)}

        ##### crop type and technology ##### 
        # separate entry for each crop type
        self.irrig = {} # only gets turned on by outgrower in LSLA
        self.farmed_fraction = np.full([self.T, self.n_plots], np.nan) # for overall fallow practices
        self.outgrower = np.full([self.T, self.n_plots], False)
        self.ha_farmed = {}
        self.fertilizer = {} # kg N/ha farmed
        for crop in self.ag_types:
            self.ha_farmed[crop] = np.full([self.T, self.n_plots], 0.)
            self.fertilizer[crop] = np.full([self.T, self.n_plots], 0) # store as int (kg N/ha)
            self.irrig[crop] = False # only gets turned on by outgrower
        # for multi-year market crops -- keep track of the area of crop in its 1st year and ready for harvest
        self.ha_mkt_yr1 = np.full([self.T, self.n_plots], 0.)
        self.ha_mkt_yr2 = np.full([self.T, self.n_plots], 0.)

        ##### crop yields #####
        self.max_yield = {'subs' : self.max_subs_yield, 'mkt' : self.max_mkt_yield}
        self.regional_crop_production = np.full(self.T, 0) # includes agents and LSLA
        self.residue_production = np.full([self.T, self.n_plots], 0) # for overall fallow practices
        self.rf_factors = np.full([self.T, self.n_plots], np.nan)
        self.labor_yield_effect = np.full([self.T, self.n_plots], np.nan) # same for both crops
        self.yields = {}
        self.yields_unconstrained = {}
        self.max_with_nutrients = {}
        self.nutrient_factors = {}
        for crop in self.ag_types:
            self.yields[crop] = np.full([self.T, self.n_plots], -9999) # kg
            self.yields_unconstrained[crop] = np.full([self.T, self.n_plots], -9999)# kg
            self.max_with_nutrients[crop] = np.full([self.T, self.n_plots], -9999)# kg
            self.nutrient_factors[crop] = np.full([self.T, self.n_plots], np.nan)
        # random effect -- init at start to control stochasticity -- same for all crop types
        # self.errors = np.random.normal(1, self.random_effect_sd, (self.T, self.n_plots))
        self.errors = np.random.lognormal(self.random_effect_mu, self.random_effect_sd, (self.T, self.n_plots)) # median : 1, mean : exp(mu+sigma^2/2)
        self.errors[self.errors < 0] = 0
        # code.interact(local=dict(globals(), **locals()))

    def update_soil(self, agents):
        '''
        simulate the evolution of the land throughout the year
        assume that the organic matter is consistent between all fields
        but the inorganic N can be different for different crops
        '''
        t = self.t[0]
        # if not self.som_dynamics:
        #     self.organic[t+1] = self.organic[t]
        #     tmp_fert = self.fertilizer_input(agents) 
        #     for crop in self.ag_types:
        #         self.inorganic[crop][t] = self.slow_mineralization_rate * self.organic[t] # soil mineralization
        #         self.inorganic[crop][t] += tmp_fert[crop]
        #     return

        ### initialize -- assume inorganic is reset each year
        organic = copy.copy(self.organic[t])
        inorganic = {}
        for crop in self.ag_types:
            inorganic[crop] = np.full(self.n_plots, 0.)

        ### mineralization: assume a linear decay model
        # assume the stocks from last year mineralize straight away
        mineralization = self.slow_mineralization_rate * organic
        organic -= mineralization
        for crop in self.ag_types:
            inorganic[crop] += mineralization

        ### agent inputs
        residue = self.crop_residue_input(agents)
        livestock = self.livestock_SOM_input(agents) # kgN/ha
        fallow = self.fallow_input(agents) # set to zero for model experiments
        fertilizer = self.fertilizer_input(agents) # separate for each crop type

        # these additions are split between organic and inorganic matter
        organic += (1-self.fast_mineralization_rate) * (residue + livestock + fallow)
        for crop in self.ag_types:
            inorganic[crop] += self.fast_mineralization_rate * (residue + livestock + fallow)
            inorganic[crop] += fertilizer[crop]

        ### constrain to be within bounds
        organic[organic < 0] = 0
        organic[organic > self.max_organic_N] = self.max_organic_N

        ### inorganic losses: loss of inorganic is a linear function of SOM
        inorg_loss_rate = (self.loss_min + (self.max_organic_N-organic)/self.max_organic_N * (self.loss_max - self.loss_min))
        for crop in self.ag_types:
            losses = inorganic[crop] * inorg_loss_rate
            inorganic[crop] -= losses
            inorganic[crop][inorganic[crop] < 0] = 0 # constrain

        ### save final values
        self.organic[t+1] = organic if self.som_dynamics else self.organic[t] # start of next year
        # code.interact(local=dict(globals(), **locals()))
        for crop in self.ag_types:
            self.inorganic[crop][t] = inorganic[crop] # end of this year (for yields)

    def crop_residue_input(self, agents):
        '''
        apply crop residues from the previous year to fields
        assume there's a conversion factor from the crop yields
        and convert back to "nitrogen"
        '''
        if self.t[0] > 0:
            tmp = np.full(self.n_plots, 0.)
            pos = agents.has_land
            # tmp[pos] = self.residue_production[self.t[0]-1,pos] / agents.land_area[pos] / self.residue_CN_conversion # kgN/ha = kgN / ha * kgN/kgC
            tmp[pos] = self.residue_production[self.t[0]-1,pos] / agents.land_area[pos] * self.residue_N_frac # kgN/ha = kg / ha * kgN/kg dry matter
            # code.interact(local=dict(globals(), **locals()))
            return tmp
        else:
            return np.full(self.n_plots, 0.)

    def livestock_SOM_input(self, agents):
        '''
        additional livestock SOM inputs come from import from rangeland
        assume 100% of nutrients from livestock grazed on rangeland are imported
        '''
        external_ls_per_ha = np.full(self.n_plots, 0.)
        pos = agents.has_land
        ls_inp = self.all_inputs['livestock']
        # agents' livestock are split equally over their land. birr / ha
        # use last year's livestock grazed on rangeland
        if self.t[0]==0:
            external_ls_per_ha = np.full(agents.N, 0)
        else:
            external_ls_per_ha[pos] = agents.herds_on_rangeland[self.t[0]-1, pos] / agents.land_area[pos]
        
        N_per_ha = external_ls_per_ha * ls_inp['N_production'] * ls_inp['frac_N_import']  # head/ha * kgN/head * __ = kgN/ha
        # code.interact(local=dict(globals(), **locals()))
        return N_per_ha

    def fertilizer_input(self, agents):
        '''
        calculate the input from fertilizer
        this amount is applied ONLY to the farmed land (i.e., not any land in conservation)
        but this has no implications for the soil calculations
        '''
        inputs = {}
        for crop in self.ag_types:
            inputs[crop] = np.full(self.n_plots, 0.)
            if crop == 'subs':
                # agents choose to apply fertilizer
                inputs[crop][agents.choices['fertilizer'][self.t[0]]==True] += self.all_inputs['adaptation']['fertilizer_application_rate']
            elif crop == 'mkt':
                inputs[crop][agents.outgrower_yr[self.t[0]]] += self.all_inputs['LSLA']['fert_amt_ratio'] * \
                                                            self.all_inputs['adaptation']['fertilizer_application_rate'] # all agents in outgrower apply fertilizer

        return inputs # kg N/ha

    def fallow_input(self, agents):
        '''
        SOM additions from fallow
        assume that legumes are grown on fallow plots
        or some other type of fixation naturally occurs
        '''
        inputs = np.full(self.n_plots, self.fallow_N_add * self.fallow_frac) # kgN/ha = kgN/ha_fallow * ha_fallow/ha
        inputs[~agents.fallow[self.t[0]]] = 0
        return inputs

    def crop_yields(self, agents, climate):
        '''
        calculate crop yields
        assume yield = (MAX_VAL * climate_reduction +/- error) * nutrient_reduction
        '''
        t = self.t[0]
        self.rf_factors[t] = self.calculate_rainfall_factor(climate.rain[t])
        for crop in self.ag_types:
            ixs = self.ha_farmed[crop][t] > 0
            if ixs.sum() > 0:
                # rainfall effect
                rain_facs = 1 if self.irrig[crop] else self.rf_factors[t,ixs]
                # nutrient unconstrained yield
                self.yields_unconstrained[crop][t,ixs] = self.max_yield[crop] * rain_facs # kg/ha
                # factor in nutrient contraints
                # self.max_with_nutrients[crop][t,ixs] = self.inorganic[crop][t,ixs] / (1/self.crop_CN_conversion+self.residue_multiplier/self.residue_CN_conversion) # kgN/ha / (kgN/kgC_yield) = kgC/ha ~= yield(perha
                self.max_with_nutrients[crop][t,ixs] = self.inorganic[crop][t,ixs] / (self.crop_N_frac+self.residue_multiplier*self.residue_N_frac) # kgN/ha / (kgN/kgC_yield) = kgC/ha ~= yield(perha
                self.yields[crop][t,ixs] = np.minimum(self.yields_unconstrained[crop][t,ixs], self.max_with_nutrients[crop][t,ixs]) * self.errors[t,ixs] # kg/ha
                with np.errstate(invalid='ignore'):
                    self.nutrient_factors[crop][t,ixs] = self.yields[crop][t,ixs] / self.yields_unconstrained[crop][t,ixs]
                    self.nutrient_factors[crop][t,ixs] = np.minimum(self.nutrient_factors[crop][t,ixs], 1)
                # labor factors
                if self.labor_yield_reduction:
                    agents.labor_density[t,ixs] = agents.hh_size[ixs] / (agents.tot_ag_labor[t] + agents.ls_labor[t] + agents.salary_labor[t])[ixs]
                    self.labor_yield_effect[t,ixs] = 1 - np.exp(-agents.labor_density[t,ixs])
                    self.yields[crop][t,ixs] = (self.yields[crop][t,ixs] * self.labor_yield_effect[t,ixs]).astype(int)
                    if self.labor_yield_effect[t].min() < 0:
                        print('LAND: negative labor effect')
                        code.interact(local=dict(globals(), **locals()))
                # harvest crops
                if (crop=='mkt') and (self.ha_farmed[crop][t,ixs].sum()>0) and (self.all_inputs['adaptation']['outgrower']['harvest_time']=='2yr'):
                    # sum the yields from the past two years... but only harvest the yield in the 2nd year (ie ha_mkt_yr2)
                    agents.crop_production[crop][t,ixs] = (self.yields[crop][t-1,ixs]+self.yields[crop][t,ixs]) * self.ha_mkt_yr2[t,ixs]
                else:
                    # attribute to agents -- ha_farmed is already adjusted for their fallowing fractions
                    agents.crop_production[crop][t,ixs] = self.yields[crop][t,ixs] * self.ha_farmed[crop][t,ixs] # harvest all that is farmed

                if self.yields[crop][t,ixs].min() < 0:
                    print('LAND: negative yield')
                    code.interact(local=dict(globals(), **locals()))


        agents.tot_crop_production[t] = np.sum(np.array([agents.crop_production[crop][t] for crop in self.ag_types]), axis=0)
        self.residue_production[t] = agents.tot_crop_production[t] * self.residue_multiplier * self.residue_loss_factor # kg total
        self.regional_crop_production[t] += agents.tot_crop_production[t].sum()
        # code.interact(local=dict(globals(), **locals()))

    def LSLA_production(self, lsla, climate):
        '''
        simulate the production (soil, climate, and everything) for the LSLA
        '''
        t = self.t[0]
        ## NUTRIENTS
        organic = lsla.organic[t]
        inorganic = 0
        mineralization = self.slow_mineralization_rate * organic
        organic -= mineralization
        inorganic += mineralization
        # residue = lsla.production[t-1] * self.residue_multiplier * self.residue_loss_factor / lsla.productive_area / self.residue_CN_conversion 
        residue = lsla.production[t-1] * self.residue_multiplier * self.residue_loss_factor / lsla.productive_area * self.residue_N_frac 
        fertilizer = self.all_inputs['LSLA']['fert_amt_ratio'] * \
                self.all_inputs['adaptation']['fertilizer_application_rate'] # assume same rate as outgrower experiment
        inorganic += self.fast_mineralization_rate * residue + fertilizer
        organic += (1-self.fast_mineralization_rate) * residue
        organic = min(max(organic, 0), self.max_organic_N) # constrain to bounds
        inorg_loss_rate = (self.loss_min + (self.max_organic_N-organic)/self.max_organic_N * (self.loss_max - self.loss_min))
        inorganic -= min(inorganic, inorg_loss_rate * inorganic) # subtract the loss
        lsla.organic[t+1] = organic if self.som_dynamics else lsla.organic[t]
        lsla.inorganic[t] = inorganic
        ## YIELDS
        # note - if they are doing 2-year crop, assume they split it such that they harvest the same amount each year
        # this shouldn't affect overall dynamics
        # rainfall: eff max and red max already calculated for crop yields of farmers
        lsla.rf_factors[t] = 1 if lsla.irrig or climate.rain[t]>self.rain_crit else self.eff_max - \
                                        (1 - lsla.organic[t:(t+1)].mean()/self.max_organic_N) * self.red_max       
        yield_unconstrained = self.max_yield['mkt'] * lsla.rf_factors[t]
        # max_with_nutrients = lsla.inorganic[t] / (1/self.crop_CN_conversion+self.residue_multiplier/self.residue_CN_conversion)
        max_with_nutrients = lsla.inorganic[t] / (self.crop_N_frac+self.residue_multiplier*self.residue_N_frac)
        yields = np.minimum(yield_unconstrained, max_with_nutrients) # no random factor here
        lsla.production[t] = yields * lsla.productive_area
        self.regional_crop_production[t] += lsla.production[t]

    def calculate_rainfall_factor(self, rain, virtual=False):
        '''
        convert the rainfall value (in 0,1) to a yield reduction factor
        '''
        if rain > self.rain_crit:
            if virtual:
                return 1
            else:
                return np.full(self.n_plots, 1) # no effect
        else:
            # organic matter reduces rainfall sensitivity
            # first, calculate value with maximum organic N
            a = self.rain_cropfail_high_SOM
            b = self.rain_cropfail_low_SOM
            c = self.rain_crit
            self.eff_max = (a-rain) / (a-c)
            # if this is a "virtual" calculation we don't account for the SOM
            if virtual:
                return max(self.eff_max, 0)
            # now, if SOM=0, how much is it reduced?
            # this is a function of the difference in the slopes of the two lines
            self.red_max = (c - rain) * (1/(c-b) - 1/(c-a))
            # now factor in the fields' actual SOM values
            # assume the average of the start and end of the year
            mean_organic = np.mean(self.organic[[self.t[0], self.t[0]+1]], axis=0)
            rf_effects = self.eff_max - (1 - mean_organic/self.max_organic_N) * self.red_max         

            return np.maximum(rf_effects, 0)