'''
add a land acquisition into the model
'''
import numpy as np
import pandas as pd
import code
import copy
import sys
from collections import Counter

class LSLA:
    def __init__(self, inputs, agents, land, rangeland, market):
        '''
        implement the LSLA "changes"
        this operates for BOTH the control and treatment households. i.e., simulate the recall data for both C and T households.
        '''
        mod_inp = inputs['model']
        # attribute the parameters to the object
        self.all_inputs = inputs
        self.inputs = inputs['LSLA']
        for key, val in self.inputs.items():
            setattr(self, key, val)
        if self.outgrower:
            for key, val in self.all_inputs['adaptation']['outgrower'].items():
                setattr(self, key, val)

        # import datasets
        lsla_df = pd.read_csv(self.info_file)
        lulc_df = pd.read_csv(mod_inp['LULC_file'], index_col=0)
        lulc_df = lulc_df[lulc_df.site==mod_inp['site']]
        # change setting if in artificial LSLA
        if mod_inp['site'] == 'ORX':
            lulc_df['sh_frac_lsla_buf'] = self.frac_smallholder
        # code.interact(local=dict(globals(), **locals()))

        # work out total area
        self.area = float((agents.land_area.sum() + rangeland.size_ha) * lulc_df['area_lsla_frac'])
        self.productive_area = self.area * self.implementation_fraction * self.LUC
        # self.productive_area = 0 if self.outgrower else self.area * self.implementation_fraction * self.LUC

        #### EMPLOYMENT ####
        # assume some additional employment in the treatment site only
        # and no additional employment if it's outgrower
        if ((mod_inp['treat'] == 'Treatment') and (self.LUC) and not (self.outgrower)):
            if self.job_data_source == 'empirical':
                self.tot_salary_jobs = self.empirical_jobs_per_agent[mod_inp['site']] * agents.N * self.implementation_fraction
            elif self.job_data_source == 'artificial':
                self.tot_salary_jobs = self.employment_per_ha * self.area * self.implementation_fraction
        else:
            self.tot_salary_jobs = 0
        market.salary_job_avail_total += round_down(self.tot_salary_jobs, market.salary_job_increment)

        #### COMMONS LAND CHANGE ####
        rangeland.size_ha_pre = copy.deepcopy(rangeland.size_ha)
        ## first, determine whether to use the treatment or control value...
        # simulate this for both control and treatment sites
        # extract the non-ag land fraction. the unit here is (ha-natural)/(ha-tot)
        col = 'nat_frac_post' if self.LUC else 'nat_frac_post_cf'
        # if (mod_inp['treat']=='Treatment' and mod_inp['matched']):
        #     # if it's a matched estimate, we take the counterfactual value (i.e., ..._cf)
        #     if (self.outgrower and self.LUC):
        #         col = 'nat_frac_post' # still lose the LSLA area even if using matched HH outcomes
        #     else:
        #         col = 'nat_frac_post_cf'
        # elif (self.outgrower and not self.LUC):
        #     # if it's outgrower with no LUC we take the counterfactual value also
        #     # (note: outgrower with LUC we assume that the LSLA land area is turned into agriculture)
        #     col = 'nat_frac_post_cf'
        # else:
        #     col = 'nat_frac_post'
        ## second, attribute this change based on the type of LUC we are modeling
        if mod_inp['commons_iLUC']:
            # in this case, model indirect changes
            val = float(lulc_df.query('treat=="{}"'.format(mod_inp['treat']))[col])
            # convert this to the same unit as used with the input data: ha-natural:ha_smallholder
            # this assumes that all non-rangeland/forest is smallholder (this is what the model assumes)
            val2 = val / (1-val)
            # attribute this change
            rangeland.size_ha = val2 * rangeland.farm_area_ha
        else:
            if col == 'nat_frac_post':
                nat_frac_lost = lulc_df.area_lsla_frac * (1 - lulc_df.sh_frac_lsla_buf) # as a fraction of the total site (smallholder + natural)
                nat_ha_lost = nat_frac_lost * (rangeland.size_ha_pre + agents.land_area.sum())
            else:
                nat_ha_lost = 0
            rangeland.size_ha = float(rangeland.size_ha_pre - nat_ha_lost)
            # code.interact(local=dict(globals(), **locals()))
        
        # account for partial implementation if it is an outgrower scheme
        if self.outgrower and self.LUC:
            dif = rangeland.size_ha_pre - rangeland.size_ha
            rangeland.size_ha = rangeland.size_ha_pre - dif * self.implementation_fraction


        #### AG-LAND CHANGE ####
        # save the pre values
        agents.land_area_pre = copy.deepcopy(agents.land_area)
        agents.has_land_pre = copy.deepcopy(agents.has_land)
        # attribute the post values
        agents.land_area = agents.land_area_post
        # if we have forced outgrower, we need to record this and not take land away from agents
        if (self.outgrower and self.LUC):
            # here, agents are forced to be in outgrower. 
            agents.forced_outgrower[agents.in_prospective_outgrower] = True
            # HHs lose (or gain) the counterfactual amount of land. 
            # if using LULC data, this is zero, if using survey data, this is the matched counterfactual HHs
            agents.land_area = copy.deepcopy(agents.land_area_post_cf)
            # reduce the LSLA implementation area by the amt that the smallholders contribute
            self.productive_area -= agents.forced_outgrower_ha.sum()
        # make final assignments
        agents.has_land = agents.land_area > 0
        agents.has_land_post = agents.has_land
        agents.land_area_change = agents.land_area - agents.land_area_pre

        #### LAND RENTAL CHANGE ####
        agents.land_rented_pre = copy.deepcopy(agents.land_rented)
        agents.land_rented = agents.land_rented_post

        #### OUTGROWER EFFECTS ####
        if self.outgrower and self.irrig:
            land.irrig['mkt'] = True
        if self.outgrower and self.join:
            agents.can_join_outgrower = True

        # simulate the honoring/breaching of contracts by the firm
        # this is only relevant in the CF_join case
        # do this ex-ante now, but agents don't see it
        market.honor_contract = np.random.binomial(n=1, p=self.all_inputs['adaptation']['outgrower']['prob_honor'], size=(agents.T, agents.N)).astype(bool)

        #### PRODUCTIVITY
        t = agents.t[0]
        self.production = np.full(agents.T, 0)
        self.organic = np.full(agents.T+1, np.nan)
        # self.organic[t] = land.organic[t].mean() # init the soil organic matter as the average over all agents now
        self.organic[t] = np.median(land.organic[t]) # init the soil organic matter as the median over all agents now.. use median b/c of the lognormal distribution
        self.inorganic = np.full(agents.T, 0.)
        self.rf_factors = np.full(agents.T, 0.)

        # checks
        if mod_inp['print_init']:
            print('  Rng col : {}'.format(col))

def round_up(amts, stepsize):
    '''
    round amts up to the nearest stepsize
    '''
    return np.ceil(amts/stepsize) * stepsize

def round_down(amts, stepsize):
    '''
    round amts up to the nearest stepsize
    '''
    return np.floor(amts/stepsize) * stepsize