'''
model class
'''
import numpy as np
import pandas as pd
from .agents import Agents
from .land import Land
from .climate import Climate
from .rangeland import Rangeland
from .market import Market
from .decision import Decision
from.lsla import LSLA
import code

class Model():
    def __init__(self, inputs, read_base_inputs=False):
        ## set the LSLA settings
        typ = inputs['LSLA']['type']
        if typ == 'none':
            inputs['LSLA']['LUC'] = False
            inputs['LSLA']['outgrower'] = False
        
        elif typ == 'displacement':
            inputs['LSLA']['LUC'] = True
            inputs['LSLA']['outgrower'] = False
        
        elif typ == 'CF_forced':
            inputs['LSLA']['LUC'] = True
            inputs['LSLA']['outgrower'] = True
            inputs['adaptation']['outgrower']['join'] = False
        
        elif typ == 'CF_join':
            inputs['LSLA']['LUC'] = False
            inputs['LSLA']['outgrower'] = True
            inputs['adaptation']['outgrower']['join'] = True
        

        # parameter exceptions
        # if inputs['adaptation']['outgrower']['fert_amt'] == 'same_as_subs':
        #     inputs['adaptation']['outgrower']['fert_amt'] = inputs['adaptation']['fertilizer_application_rate']
        # attribute the parameters to the object
        self.all_inputs = inputs
        self.inputs = inputs['model']
        for key, val in self.inputs.items():
            setattr(self, key, val)

        np.random.seed(self.seed)

        # initialize the sub-objects
        self.climate = Climate(inputs)
        self.market = Market(inputs)
        self.agents = Agents(inputs, self.market)
        self.land = Land(self.agents, inputs)
        self.rangeland = Rangeland(np.sum(self.agents.land_area), inputs)
        # set the time
        # save as list so it is mutable (stays same over all objects)
        self.t = self.agents.t = self.land.t = self.climate.t = self.rangeland.t = [0] 
        self.lsla_active = False
        self.adap_properties = {
            'adap' : False,
            'type' : 'none'
        }
        np.random.seed(self.seed*2) # set seed again -- init adap option uses random numbers only if adap option selected

        if self.lsla_experiment:
            # import the LSLA info
            df = pd.read_csv(inputs['LSLA']['info_file'])
            self.t_lsla = np.array(df.loc[df.site==self.site,'start'])[0] - self.year_start

    def step(self):
        '''
        advance the simulation by one year
        '''
        if (self.lsla_experiment) and (self.t[0]==self.t_lsla):
            self.lsla = LSLA(self.all_inputs, self.agents, self.land, self.rangeland, self.market) # implement the LSLA "changes" (control and treatment)
        Decision.annual_decisions(self.agents, self.land, self.market)
        self.market.allocate_salary_labor(self.agents)
        self.land.update_soil(self.agents)
        self.land.crop_yields(self.agents, self.climate)
        if (self.lsla_experiment) and (self.t[0]>=self.t_lsla) and (self.lsla.productive_area > 0):
            self.land.LSLA_production(self.lsla, self.climate)            
        self.agents.income_and_food_consumption(self.market, self.climate, self.land, self.adap_properties, False) # false means estimate==False. #self.land, self.climate, self.market)
        self.rangeland.update(self.climate, self.agents, self.land) # includes livestock reproduction
        self.agents.coping_measures(self.land, self.market)
        self.agents.livestock_stocking_calculation(self.land, self.rangeland, self.market)
        self.agents.blf.update(self.agents, self.land, self.market)
        self.rangeland.size_ha_t[self.t[0]] = self.rangeland.size_ha
        # save for next year
        self.t[0] += 1 
        self.agents.livestock[self.t[0]] = self.agents.ls_obj
        # if not self.agents.savings_acct:
        #     self.agents.savings[self.t[0]] = 0 # agents cannot carry over money to the next year
        self.agents.wealth[self.t[0]] = self.agents.ls_obj*self.market.livestock_cost + self.agents.savings[self.t[0]]     