import numpy as np
import pandas as pd
import code
import copy
from collections import Counter

class Rangeland():
    def __init__(self, farm_area_ha, inputs):
        # attribute the parameters to the object
        self.all_inputs = inputs
        self.inputs = inputs['rangeland']
        for key, val in self.inputs.items():
            setattr(self, key, val)
        mod_inp = self.all_inputs['model']
        self.T = mod_inp['T']
        self.farm_area_ha = farm_area_ha

        # calculate size (in ha)
        if 'rangeland' in mod_inp['read_from_file']:
            df = pd.read_csv(mod_inp['LULC_file'], index_col=0)
            # fraction = float(df.query('site=="{}" & treat=="{}"'.format(mod_inp['site'], mod_inp['treat']))['nat_frac_pre']) # unit: ha-nat/ha-tot
            fraction = 1 - float(df.query('site=="{}" & treat=="{}"'.format(mod_inp['site'], mod_inp['treat']))['sh_frac_buf_pre']) # unit: ha-nat/ha-tot
            self.range_farm_ratio = fraction / (1-fraction) # unit: ha-nat/ha-smallholder
        
        self.size_ha = self.range_farm_ratio * farm_area_ha
        self.size_ha_t = np.full(self.T, np.nan)
        # initialize arrays
        # note: this represents the state at the START of the time step
        self.G = np.full(self.T+1, -99) # store as integer (kg/ha)
        self.R = np.full(self.T+1, -99) # kg/ha
        self.F_avail = np.full(self.T, -99)
        self.G_no_cons = np.full(self.T, -99) # store as integer (kg/ha)
        self.destocking_rqd = np.full(self.T, False)
        val = -99 if self.integer_consumption else np.nan
        self.demand_intensity = np.full(self.T, val)
        self.livestock_supported = np.full(self.T, val)
        self.R[0] = 0 # reserve biomass. assume zero
        self.G[0] = self.G_constant

        # checks
        if mod_inp['print_init']:
            print('  Rng : {}_{}'.format(mod_inp['site'], mod_inp['treat']))

    def update(self, climate, agents, land):
        '''
        simulate the rangeland dynamics
        This model is copied from Gunnar
        '''
        t = self.t[0]
        herds = agents.ls_obj

        ## 1. livestock reproduction
        if not self.reproduce_at_end:
            self.reproduction(herds, agents, t)

        # rangeland dynamics (not modeled so assume constant)
        self.G[t+1] = self.G_constant
        self.G_no_cons[t] = copy.deepcopy(self.G[t+1])
        self.R[t+1] = 0

        ## 3. livestock consumption and destocking
        destocking_total = self.consumption(agents, land, herds, t)
        herds_pre = copy.deepcopy(agents.herds_on_rangeland[t])
        agents.attempted_herds_on_rangeland[t] = copy.deepcopy(herds_pre)
        self.apportion_destocking(destocking_total, agents.herds_on_rangeland[t])
        agents.destocking_rqd_rangeland[t] = herds_pre > agents.herds_on_rangeland[t]
        # add together the livestock on residue and rangeland
        # if there's been a "partial destocking" required on the rangeland, the whole animal must be destocked
        # hence, convert to integer. this takes the "floor" of each number, so first round to closest decimal in case of rounding errors
        agents.ls_obj = np.round(agents.herds_on_residue[t] + agents.herds_on_rangeland[t] + 
            agents.herds_on_others_residue[t], 1).astype(int)
        # code.interact(local=dict(globals(), **locals()))
        agents.ls_rangeland[t] = copy.deepcopy(agents.ls_obj)

        # 5. reproduction if at end
        if self.reproduce_at_end:
            self.reproduction(agents.ls_obj, agents, t)

        if np.sum(agents.ls_obj<0)>0:
            print('ERROR: negative livestock in rangeland.update()')
            code.interact(local=dict(globals(), **locals())) 

    def reproduction(self, herds, agents, t):
        # note: we model livestock as integer valued
        # each animal has the same probability of reproducing
        # control stochasticity b/c we have variable number of livestock
        birth_rate = self.all_inputs['livestock']['birth_rate']
        rndm_num = np.random.randint(1e6)
        if birth_rate > 0:
            births = np.random.binomial(herds, birth_rate)
            herds += births
    
        np.random.seed(rndm_num)
        agents.ls_reprod[t] = copy.deepcopy(herds)

    def consumption(self, agents, land, herds, t):
        '''
        simulate the consumption of on-farm crop residues
        as well as green and reserve biomass
        this function operates with dynamic rangeland (ie rangeland_dynamics==T)
        as well as with rangeland_dynamics==T and constant_fraction_livestock==F
        livestock consumption is achieved via a mix of
        1. on-farm residues
        2. others' on-farm residues
        3. the communal rangeland
        '''
        ls_consumption = self.all_inputs['livestock']['consumption']
        ## 1. own crop residues
        agents.herds_on_residue[t] = np.minimum(herds, land.residue_production[t] / ls_consumption) # kg / (kg/head) = head
        if self.integer_consumption:
            agents.herds_on_residue[t] = np.round(agents.herds_on_residue[t], 1).astype(int)
        
        ## 2. sharing of residues
        if self.all_inputs['livestock']['crop_residue_sharing']:
            leftover_herds = herds - agents.herds_on_residue[t]
            leftover_residues = land.residue_production[t] - agents.herds_on_residue[t] * ls_consumption
            agents.herds_on_others_residue[t] = self.residue_sharing(agents, leftover_herds, leftover_residues, ls_consumption)
        else:
            agents.herds_on_others_residue[t] = 0

        ## 3. rangeland
        # agents.herds_on_residue[t] = np.minimum(herds, land.residue_production[t] / self.all_inputs['livestock']['consumption'])
        # ^ take the floor of this. keep as integer
        # demand for the rangeland
        agents.herds_on_rangeland[t] = np.round(herds - agents.herds_on_residue[t] - agents.herds_on_others_residue[t], 3) # some rounding errors cause -ve herd sizes...

        if self.size_ha == 0:
            # clause if there is NO rangeland left (due to LSLA)
            self.demand_intensity[t] = np.inf
            self.G[t+1] = 0
            self.R[t] = 0
            destocking_total = int(np.ceil(agents.herds_on_rangeland[t].sum()))
        else:
            self.demand_intensity[t] = np.sum(agents.herds_on_rangeland[t]) * ls_consumption / self.size_ha # unit = kg/ha
            # how is this demand satisfied...?
            if self.G[t+1] > self.demand_intensity[t]:
                # green biomass satisfies demand
                self.G[t+1] -= self.demand_intensity[t]
                destocking_total = 0 # no destocking rqd
            else:
                # need to use reserve biomass
                # note: this is zero if rangeland_dynamics==F
                reserve_demand = self.demand_intensity[t] - self.G[t+1]
                reserve_limit = 0 if self.R[t] == 0 else self.gr2 * self.R[t]
                self.G[t+1] = 0
                if reserve_demand < reserve_limit:
                    # reserve can supply the demand
                    self.R[t] -= reserve_demand
                    destocking_total = 0 # no destocking rqd
                else:
                    # still remaining shortfall
                    self.R[t] -= reserve_limit
                    deficit = reserve_demand - reserve_limit
                    destocking_total = int(np.ceil(deficit * self.size_ha / ls_consumption)) # kg/ha * ha / (kg/head) = head (must be integer)

        self.destocking_rqd[t] = True if np.sum(destocking_total)>0 else False
        self.livestock_supported[t] = np.sum(agents.herds_on_rangeland[t])-destocking_total

        if np.sum(agents.herds_on_residue[t]<0)>0:
            print('ERROR: negative herds on residue')
            code.interact(local=dict(globals(), **locals())) 

        return destocking_total

    def residue_sharing(self, agents, herd, resids, ls_consumption):
        '''
        model the sharing of leftover residues between agents' neighbors
        herd=herds not fed by own residues, resids=leftover residuals after own livestock feeding
        assume random order of consumption
        '''
        rand_int = np.random.randint(1e6)
        ls_fed = np.full(agents.N, 0.)
        rndm_order = np.random.choice(agents.N, replace=False, size=agents.N)
        np.random.seed(rand_int)
        for ag in rndm_order: # need to use for loop .... :/
            if herd[ag] > 0:
                resid_access = resids[agents.network[ag]]
                resid_cumsum = resid_access.cumsum()
                cons = min(herd[ag]*ls_consumption, resid_cumsum[-1])
                resid_post_feeding = resid_cumsum - cons # subtract consumption
                resid_post_feeding[resid_post_feeding<0] = 0 # correct negative values -- these ppl have nothing left
                resid_post_feeding[1:] -= resid_post_feeding[:-1] # remove the cumsum
                # attribute these changes
                ls_fed[ag] = cons/ls_consumption
                resids[agents.network[ag]] = resid_post_feeding

        return ls_fed

    def apportion_destocking(self, total, range_herds):
        '''
        apportion destocking randomly between agents
        each livestock has an equal probability of being destocked
        NOTE: variable number of calls to np.random so control stochasticity
        NOTE2: agents receive no income for this destocking !!
        '''
        rand_int = np.random.randint(1e6)
        if total > 0:
            tot_ls = np.sum(range_herds)
            total = np.ceil(min(total, tot_ls)).astype(int) # ceiling -- partial livestock can be destocked also
            
            if self.integer_consumption:
                destock_ix = np.random.choice(np.arange(tot_ls), size=total, replace=False) # indexes of livestock
                owner_ix = np.repeat(np.arange(range_herds.shape[0]), range_herds)
                owner_destocks = owner_ix[destock_ix]
                destock_counts = np.array(list(Counter(owner_destocks).items()))
            else:
                # create a list of all of the livestock (even partial livestock)
                # partial livestock have a lower probability of being selected for destocking
                rounded = np.ceil(range_herds).astype(int)
                owner_ix = np.repeat(np.arange(rounded.shape[0]), rounded)
                partial_ls_amt = range_herds - np.floor(range_herds)
                partial_ixs = np.cumsum(rounded[rounded>0]) - 1 # only consider the agents that have livestock
                probs = np.full(owner_ix.shape, 1.) # probability of being selected for destocking
                # the partial ixs includes those with _whole_ numbers of livestock (e.g., 5.)
                # we want to keep these ixs as a probability of 1, so filter also on partial_ls_amt>0
                probs[partial_ixs[partial_ls_amt[rounded>0]>0]] = partial_ls_amt[(rounded>0) * (partial_ls_amt>0)]
                # select the owner ids of the livestock that need to be destocked
                destock_ixs = np.random.choice(owner_ix.shape[0], size=total, replace=False, p=probs/probs.sum())
                owner_destocks = owner_ix[destock_ixs]
                destock_amts = probs[destock_ixs] # some of the destocks can be partial
                c = Counter()
                for o, ownr_id in enumerate(owner_destocks):
                    c.update({ownr_id:destock_amts[o]})
                destock_counts = np.array(list(c.items()))

            range_herds[destock_counts[:,0].astype(int)] -= destock_counts[:,1]

        np.random.seed(rand_int)